#!python3

"""
FSA DICTIONARY BUILDER

by Olivier R.
License: MPL 2

This tool encodes lexicon into an indexable binary dictionary
Input files MUST be encoded in UTF-8.
"""

import sys
import os
import collections
import json
import time
import re
import traceback

from . import str_transform as st
from .progressbar import ProgressBar



dLexiconData = {}

def readFile (spf):
    "generator: read file <spf> and return for each line a list of elements separated by a tabulation."
    print(" < Read lexicon: " + spf)
    if os.path.isfile(spf):
        dLexiconData.clear()
        with open(spf, "r", encoding="utf-8") as hSrc:
            for sLine in hSrc:
                sLine = sLine.strip()
                if sLine.startswith("##") :
                    m = re.match("## *(\\w+) *:(.*)$", sLine)
                    if m:
                        dLexiconData[m.group(1)] = m.group(2).strip()
                elif sLine and not sLine.startswith("#"):
                    yield sLine.split("\t")
        if dLexiconData:
            print("Data from dictionary:")
            print(dLexiconData)
    else:
        raise OSError("# Error. File not found or not loadable: " + spf)



class DAWG:
    """DIRECT ACYCLIC WORD GRAPH"""
    # This code is inspired from Steve Hanov’s DAWG, 2011. (http://stevehanov.ca/blog/index.php?id=115)
    # We store suffix/affix codes and tags within the graph after the “real” word.
    # A word is a list of numbers [ c1, c2, c3 . . . cN, iAffix, iTags]
    # Each arc is an index in self.lArcVal, where are stored characters, suffix/affix codes for stemming and tags.
    # Important: As usual, the last node (after ‘iTags’) is tagged final, AND the node after ‘cN’ is ALSO tagged final.

    def __init__ (self, src, cStemming, sLangCode, sLangName="", sDicName="", sDescription="", sSelectFilterRegex=""):
        print("===== Direct Acyclic Word Graph - Minimal Acyclic Finite State Automaton =====")
        cStemming = cStemming.upper()
        if cStemming == "A":
            funcStemmingGen = st.defineAffixCode
        elif cStemming == "S":
            funcStemmingGen = st.defineSuffixCode
        elif cStemming == "N":
            funcStemmingGen = st.noStemming
        else:
            raise ValueError("# Error. Unknown stemming code: {}".format(cStemming))

        aEntry = set()
        lChar = ['']; dChar = {}; nChar = 1; dCharOccur = {}
        lAff  = [];   dAff  = {}; nAff  = 0; dAffOccur = {}
        lTag  = [];   dTag  = {}; nTag  = 0; dTagOccur = {}

        self.a2grams = set()

        try:
            zFilter = re.compile(sSelectFilterRegex)  if sSelectFilterRegex  else None
        except re.error:
            print("# Error. Wrong filter regex. Filter ignored: ", zFilter)
            traceback.print_exc()
            zFilter = None

        # read lexicon
        if isinstance(src, str):
            iterable = readFile(src)
        else:
            iterable = src
        for sFlex, sStem, sTag in iterable:
            if not zFilter or zFilter.search(sTag):
                self.a2grams.update(st.getNgrams(sFlex))
                addWordToCharDict(sFlex)
                # chars
                for c in sFlex:
                    if c not in dChar:
                        dChar[c] = nChar
                        lChar.append(c)
                        nChar += 1
                    dCharOccur[c] = dCharOccur.get(c, 0) + 1
                # affixes to find stem from flexion
                sAff = funcStemmingGen(sFlex, sStem)
                if sAff not in dAff:
                    dAff[sAff] = nAff
                    lAff.append(sAff)
                    nAff += 1
                dAffOccur[sAff] = dAffOccur.get(sAff, 0) + 1
                # tags
                if sTag not in dTag:
                    dTag[sTag] = nTag
                    lTag.append(sTag)
                    nTag += 1
                dTagOccur[sTag] = dTagOccur.get(sTag, 0) + 1
                aEntry.add((sFlex, dAff[sAff], dTag[sTag]))
        if not aEntry:
            raise ValueError("# Error. Empty lexicon")

        # Preparing DAWG
        print(" > Preparing list of words")
        print(" Filter: " + (sSelectFilterRegex or "[None]"))
        lVal = lChar + lAff + lTag
        lWord = [ [dChar[c] for c in sFlex] + [iAff+nChar] + [iTag+nChar+nAff]  for sFlex, iAff, iTag in aEntry ]
        aEntry = None

        # Dictionary of arc values occurrency, to sort arcs of each node
        dValOccur = dict( [ (dChar[c], dCharOccur[c])  for c in dChar ] \
                        + [ (dAff[aff]+nChar, dAffOccur[aff]) for aff in dAff ] \
                        + [ (dTag[tag]+nChar+nAff, dTagOccur[tag]) for tag in dTag ] )

        self.sFileName = src  if isinstance(src, str)  else "[None]"
        self.sLangCode = sLangCode
        self.sLangName = sLangName
        self.sDicName = sDicName
        self.sDescription = sDescription
        if dLexiconData:
            self.sLangCode = dLexiconData.get("LangCode", self.sLangCode)
            self.sLangName = dLexiconData.get("LangName", self.sLangName)
            self.sDicName = dLexiconData.get("DicName", self.sDicName)
            self.sDescription = dLexiconData.get("Description", self.sDescription)
        self.nEntry = len(lWord)
        self.aPreviousEntry = []
        DawgNode.resetNextId()
        self.oRoot = DawgNode()
        self.lUncheckedNodes = []  # list of nodes that have not been checked for duplication.
        self.lMinimizedNodes = {}  # list of unique nodes that have been checked for duplication.
        self.lSortedNodes = []     # version 2 and 3
        self.nNode = 0
        self.nArc = 0
        self.dChar = dChar
        self.nChar = len(dChar)
        self.nAff = nAff
        self.lArcVal = lVal
        self.nArcVal = len(lVal)
        self.nTag = self.nArcVal - self.nChar - nAff
        self.cStemming = cStemming
        if cStemming == "A":
            self.funcStemming = st.changeWordWithAffixCode
        elif cStemming == "S":
            self.funcStemming = st.changeWordWithSuffixCode
        else:
            self.funcStemming = st.noStemming

        # calculated later
        self.nBytesNodeAddress = 1
        self.nBytesArc = 0
        self.nBytesOffset = 0
        self.nMaxOffset = 0

        # binary dictionary
        self.byDic = b""
        self.lByDic = []

        # build
        lWord.sort()
        oProgBar = ProgressBar(0, len(lWord))
        for aEntry in lWord:
            self.insert(aEntry)
            oProgBar.increment(1)
        oProgBar.done()
        self.finish()
        self.countNodes()
        self.countArcs()
        self.sortNodes()         # version 2 and 3
        self.sortNodeArcs(dValOccur)
        #self.sortNodeArcs2 (self.oRoot, "")
        self.displayInfo()

    # BUILD DAWG
    def insert (self, aEntry):
        "insert a new entry (insertion must be made in alphabetical order)."
        if aEntry < self.aPreviousEntry:
            sys.exit("# Error: Words must be inserted in alphabetical order.")

        # find common prefix between word and previous word
        nCommonPrefix = 0
        for i in range(min(len(aEntry), len(self.aPreviousEntry))):
            if aEntry[i] != self.aPreviousEntry[i]:
                break
            nCommonPrefix += 1

        # Check the lUncheckedNodes for redundant nodes, proceeding from last
        # one down to the common prefix size. Then truncate the list at that point.
        self._minimize(nCommonPrefix)

        # add the suffix, starting from the correct node mid-way through the graph
        if not self.lUncheckedNodes:
            oNode = self.oRoot
        else:
            oNode = self.lUncheckedNodes[-1][2]

        iChar = nCommonPrefix
        for c in aEntry[nCommonPrefix:]:
            oNextNode = DawgNode()
            oNode.arcs[c] = oNextNode
            self.lUncheckedNodes.append((oNode, c, oNextNode))
            if iChar == (len(aEntry) - 2):
                oNode.final = True
            iChar += 1
            oNode = oNextNode
        oNode.final = True
        self.aPreviousEntry = aEntry

    def finish (self):
        "minimize unchecked nodes"
        self._minimize(0)

    def _minimize (self, downTo):
        # proceed from the leaf up to a certain point
        for i in range( len(self.lUncheckedNodes)-1, downTo-1, -1 ):
            oNode, char, oChildNode = self.lUncheckedNodes[i]
            if oChildNode in self.lMinimizedNodes:
                # replace the child with the previously encountered one
                oNode.arcs[char] = self.lMinimizedNodes[oChildNode]
            else:
                # add the state to the minimized nodes.
                self.lMinimizedNodes[oChildNode] = oChildNode
            self.lUncheckedNodes.pop()

    def countNodes (self):
        "count the number of nodes of the whole word graph"
        self.nNode = len(self.lMinimizedNodes)

    def countArcs (self):
        "count the number of arcs in the whole word graph"
        self.nArc = len(self.oRoot.arcs)
        for oNode in self.lMinimizedNodes:
            self.nArc += len(oNode.arcs)

    def sortNodeArcs (self, dValOccur):
        "sort arcs of each node according to <dValOccur>"
        print(" > Sort node arcs")
        self.oRoot.sortArcs(dValOccur)
        for oNode in self.lMinimizedNodes:
            oNode.sortArcs(dValOccur)

    def sortNodeArcs2 (self, oNode, cPrevious=""):
        "sort arcs of each node depending on the previous char"
        # recursive function
        dCharOccur = getCharOrderAfterChar(cPrevious)
        if dCharOccur:
            oNode.sortArcs2(dCharOccur, self.lArcVal)
        for nArcVal, oNextNode in oNode.arcs.items():
            self.sortNodeArcs2(oNextNode, self.lArcVal[nArcVal])

    def sortNodes (self):
        "sort nodes"
        print(" > Sort nodes")
        for oNode in self.oRoot.arcs.values():
            self._parseNodes(oNode)

    def _parseNodes (self, oNode):
        # Warning: recursive method
        if oNode.pos > 0:
            return
        oNode.setPos()
        self.lSortedNodes.append(oNode)
        for oNextNode in oNode.arcs.values():
            self._parseNodes(oNextNode)

    def lookup (self, sWord):
        "return True if <sWord> is within the word graph (debugging)"
        oNode = self.oRoot
        for c in sWord:
            if self.dChar.get(c, '') not in oNode.arcs:
                return False
            oNode = oNode.arcs[self.dChar[c]]
        return oNode.final

    def morph (self, sWord):
        "return a string of the morphologies of <sWord> (debugging)"
        oNode = self.oRoot
        for c in sWord:
            if self.dChar.get(c, '') not in oNode.arcs:
                return ''
            oNode = oNode.arcs[self.dChar[c]]
        if oNode.final:
            s = "* "
            for arc in oNode.arcs:
                if arc >= self.nChar:
                    s += " [" + self.funcStemming(sWord, self.lArcVal[arc])
                    oNode2 = oNode.arcs[arc]
                    for arc2 in oNode2.arcs:
                        s += " / " + self.lArcVal[arc2]
                    s += "]"
            return s
        return ''

    def displayInfo (self):
        "display informations about the word graph"
        print(" * {:<12} {:>16,}".format("Entries:", self.nEntry))
        print(" * {:<12} {:>16,}".format("Characters:", self.nChar))
        print(" * {:<12} {:>16,}".format("Affixes:", self.nAff))
        print(" * {:<12} {:>16,}".format("Tags:", self.nTag))
        print(" * {:<12} {:>16,}".format("Arc values:", self.nArcVal))
        print(" * {:<12} {:>16,}".format("Nodes:", self.nNode))
        print(" * {:<12} {:>16,}".format("Arcs:", self.nArc))
        print(" * {:<12} {:>16,}".format("2grams:", len(self.a2grams)))
        print(" * {:<12} {:>16}".format("Stemming:", self.cStemming + "FX"))

    def getArcStats (self):
        "return a string with statistics about nodes and arcs"
        d = {}
        for oNode in self.lMinimizedNodes:
            n = len(oNode.arcs)
            d[n] = d.get(n, 0) + 1
        s = " * Nodes:\n"
        for n in d:
            s = s + " {:>9} nodes have {:>3} arcs\n".format(d[n], n)
        return s

    def writeInfo (self, sPathFile):
        "write informations in file <sPathFile>"
        print(" > Write informations")
        with open(sPathFile, 'w', encoding='utf-8', newline="\n") as hDst:
            hDst.write(self.getArcStats())
            hDst.write("\n * Values:\n")
            for i, s in enumerate(self.lArcVal):
                hDst.write(" {:>6}. {}\n".format(i, s))
            hDst.close()

    def select (self, sPattern=""):
        "generator: returns all entries which morphology fits <sPattern>"
        zPattern = None
        if sPattern:
            try:
                zPattern = re.compile(sPattern)
            except re.error:
                print("# Error in regex pattern")
                traceback.print_exc()
        yield from self._select(zPattern, self.oRoot, "")

    def _select (self, zPattern, oNode, sWord):
        # recursive generator
        for nVal, oNextNode in oNode.arcs.items():
            if nVal <= self.nChar:
                # simple character
                yield from self._select(zPattern, oNextNode, sWord + self.lArcVal[nVal])
            else:
                sEntry = sWord + "\t" + self.funcStemming(sWord, self.lArcVal[nVal])
                for nMorphVal, _ in oNextNode.arcs.items():
                    if not zPattern or zPattern.search(self.lArcVal[nMorphVal]):
                        yield sEntry + "\t" + self.lArcVal[nMorphVal]


    # BINARY CONVERSION
    def _calculateBinary (self):
        print(" > Write DAWG as an indexable binary dictionary")
        self.nBytesArc = ( (self.nArcVal.bit_length() + 2) // 8 ) + 1   # We add 2 bits. See DawgNode.convToBytes()
        self.nBytesOffset = 0
        self._calcNumBytesNodeAddress()
        self._calcNodesAddress()
        self.byDic = self.oRoot.convToBytes(self.nBytesArc, self.nBytesNodeAddress)
        for oNode in self.lMinimizedNodes:
            self.byDic += oNode.convToBytes(self.nBytesArc, self.nBytesNodeAddress)
        print("   Arc values (chars, affixes and tags): {}  ->  {} bytes".format( self.nArcVal, len("\t".join(self.lArcVal).encode("utf-8")) ))
        print("   Arc size: {} bytes, Address size: {} bytes   ->   {} * {} = {} bytes".format( self.nBytesArc, self.nBytesNodeAddress, \
                                                                                                self.nBytesArc+self.nBytesNodeAddress, self.nArc, \
                                                                                                (self.nBytesArc+self.nBytesNodeAddress)*self.nArc ))

    def _calcNumBytesNodeAddress (self):
        "how many bytes needed to store all nodes/arcs in the binary dictionary"
        self.nBytesNodeAddress = 1
        while ((self.nBytesArc + self.nBytesNodeAddress) * self.nArc) > (2 ** (self.nBytesNodeAddress * 8)):
            self.nBytesNodeAddress += 1

    def _calcNodesAddress (self):
        nBytesNode = self.nBytesArc + self.nBytesNodeAddress
        iAddr = len(self.oRoot.arcs) * nBytesNode
        for oNode in self.lMinimizedNodes:
            oNode.addr = iAddr
            iAddr += max(len(oNode.arcs), 1) * nBytesNode

    def _binaryToList (self):
        """
        Convert binary string to binary list
        BEFORE: Arc                 Address                                 Arc                 Address
                ||||||||| ||||||||| ||||||||| ||||||||| ||||||||| ||||||||| ||||||||| ||||||||| ||||||||| ||||||||| ||||||||| ||||||||| ...

        AFTER:  list of integers: [ arc, address, arc, address, arc, address, ... arc, address ]
        """
        self.lByDic = []
        nAcc = 0
        byBuffer = b""
        nDivisor = (self.nBytesArc + self.nBytesNodeAddress) / 2
        for i in range(0, len(self.byDic)):
            byBuffer += self.byDic[i:i+1]
            if nAcc == (self.nBytesArc - 1):
                self.lByDic.append(int.from_bytes(byBuffer, byteorder="big"))
                byBuffer = b""
            elif nAcc == (self.nBytesArc + self.nBytesNodeAddress - 1):
                self.lByDic.append(round(int.from_bytes(byBuffer, byteorder="big") / nDivisor))
                byBuffer = b""
                nAcc = -1
            nAcc = nAcc + 1

    def getBinaryAsJSON (self):
        "return a JSON string containing all necessary data of the dictionary (compressed as a binary string)"
        self._calculateBinary()
        self._binaryToList()
        return {
            "sHeader": "/grammalecte-fsa/",
            "sLangCode": self.sLangCode,
            "sLangName": self.sLangName,
            "sDicName": self.sDicName,
            "sDescription": self.sDescription,
            "sFileName": self.sFileName,
            "sDate": self._getDate(),
            "nEntry": self.nEntry,
            "nChar": self.nChar,
            "nAff": self.nAff,
            "nTag": self.nTag,
            "cStemming": self.cStemming,
            "dChar": self.dChar,
            "nNode": self.nNode,
            "nArc": self.nArc,
            "nArcVal": self.nArcVal,
            "lArcVal": self.lArcVal,
            "nBytesArc": self.nBytesArc,
            "nBytesNodeAddress": self.nBytesNodeAddress,
            "nBytesOffset": self.nBytesOffset,
            # Mozilla’s JS parser don’t like file bigger than 4 Mb!
            # So, if necessary, we use an hexadecimal string, that we will convert later in Firefox’s extension.
            # https://github.com/mozilla/addons-linter/issues/1361
            #"sByDic": self.byDic.hex(),
            "lByDic": self.lByDic,
            "l2grams": list(self.a2grams)
        }

    def writeAsJSObject (self, spfDst):
        "write a file (JSON or JS module) with all the necessary data"
        if not spfDst.endswith(".json"):
            spfDst += ".json"
        with open(spfDst, "w", encoding="utf-8", newline="\n") as hDst:
            hDst.write( json.dumps(self.getBinaryAsJSON(), ensure_ascii=False) )

    def _getDate (self):
        return time.strftime("%Y-%m-%d %H:%M:%S")

    def _writeNodes (self, sPathFile):
        "for debugging only"
        print(" > Write nodes")
        with open(sPathFile+".nodes.txt", 'w', encoding='utf-8', newline="\n") as hDst:
            hDst.write(self.oRoot.getTxtRepr(self.nBytesArc, self.lArcVal)+"\n")
            #hDst.write( ''.join( [ "%02X " %  z  for z in self.oRoot.convToBytes(self.nBytesArc, self.nBytesNodeAddress) ] ).strip() )
            for oNode in self.lMinimizedNodes:
                hDst.write(oNode.getTxtRepr(self.nBytesArc, self.lArcVal)+"\n")



class DawgNode:
    """Node of the word graph"""

    NextId = 0
    NextPos = 1 # (version 2)

    def __init__ (self):
        self.i = DawgNode.NextId
        DawgNode.NextId += 1
        self.final = False
        self.arcs = {}          # key: arc value; value: a node
        self.addr = 0           # address in the binary dictionary
        self.pos = 0            # position in the binary dictionary (version 2)

    @classmethod
    def resetNextId (cls):
        "set NextId to 0 "
        cls.NextId = 0

    def setPos (self): # version 2
        "define a position for node (version 2)"
        self.pos = DawgNode.NextPos
        DawgNode.NextPos += 1

    def __str__ (self):
        s = "Node " + str(self.i) + " @ " + str(self.addr) + (" [final]" if self.final else "") + "\n"
        for arc, node in self.arcs.items():
            s += "  " +str(arc)
            s += " > " + str(node.i)
            s += " @ " + str(node.addr) + "\n"
        return s

    def __repr__ (self):
        # Caution! this function is used for hashing and comparison!
        sFinalChar = "1"  if self.final  else "0"
        l = [sFinalChar]
        for arc, node in self.arcs.items():
            l.append(str(arc))
            l.append(str(node.i))
        return "_".join(l)

    def __hash__ (self):
        # Used as a key in a python dictionary.
        return self.__repr__().__hash__()

    def __eq__ (self, other):
        # Used as a key in a python dictionary.
        # Nodes are equivalent if they have identical arcs, and each identical arc leads to identical states.
        return self.__repr__() == other.__repr__()

    def sortArcs (self, dValOccur):
        "sort arcs of node according to <dValOccur>"
        self.arcs = collections.OrderedDict(sorted(self.arcs.items(), key=lambda t: dValOccur.get(t[0], 0), reverse=True))

    def sortArcs2 (self, dValOccur, lArcVal):
        "sort arcs of each node depending on the previous char"
        self.arcs = collections.OrderedDict(sorted(self.arcs.items(), key=lambda t: dValOccur.get(lArcVal[t[0]], 0), reverse=True))

    # VERSION 1 =====================================================================================================
    def convToBytes (self, nBytesArc, nBytesNodeAddress):
        """
        Convert to bytes (method 1).

        Node scheme:
        - Arc length is defined by nBytesArc
        - Address length is defined by nBytesNodeAddress

        |                Arc                |                         Address of next node                          |
        |                                   |                                                                       |
         ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓
         ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃
         ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛
         [...]
         ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓ ┏━━━━━━━━━━━━━━━┓
         ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃ ┃
         ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━┛
          ^ ^
          ┃ ┃
          ┃ ┃
          ┃ ┗━━━ if 1, last arc of this node
          ┗━━━━━ if 1, this node is final (only on the first arc)
        """
        nArc = len(self.arcs)
        nFinalNodeMask = 1 << ((nBytesArc*8)-1)
        nFinalArcMask = 1 << ((nBytesArc*8)-2)
        if not nArc:
            val = nFinalNodeMask | nFinalArcMask
            by = val.to_bytes(nBytesArc, byteorder='big')
            by += (0).to_bytes(nBytesNodeAddress, byteorder='big')
            return by
        by = b""
        for i, arc in enumerate(self.arcs, 1):
            val = arc
            if i == 1 and self.final:
                val = val | nFinalNodeMask
            if i == nArc:
                val = val | nFinalArcMask
            by += val.to_bytes(nBytesArc, byteorder='big')
            by += self.arcs[arc].addr.to_bytes(nBytesNodeAddress, byteorder='big')
        return by

    def getTxtRepr (self, nBytesArc, lVal):
        "return representation as string of node (method 1)"
        nArc = len(self.arcs)
        nFinalNodeMask = 1 << ((nBytesArc*8)-1)
        nFinalArcMask = 1 << ((nBytesArc*8)-2)
        s = "i{:_>10} -- #{:_>10}\n".format(self.i, self.addr)
        if not nArc:
            s += "  {:<20}  {:0>16}  i{:_>10}   #{:_>10}\n".format("", bin(nFinalNodeMask | nFinalArcMask)[2:], "0", "0")
            return s
        for i, arc in enumerate(self.arcs, 1):
            val = arc
            if i == 1 and self.final:
                val = val | nFinalNodeMask
            if i == nArc:
                val = val | nFinalArcMask
            s += "  {:<20}  {:0>16}  i{:_>10}   #{:_>10}\n".format(lVal[arc], bin(val)[2:], self.arcs[arc].i, self.arcs[arc].addr)
        return s


# Another attempt to sort node arcs

_dCharOrder = {
    # key: previous char, value: dictionary of chars {c: nValue}
    "": {}
}


def addWordToCharDict (sWord):
    "for each character of <sWord>, count how many times it appears after the previous character, and store result in a <_dCharOrder>"
    cPrevious = ""
    for cChar in sWord:
        if cPrevious not in _dCharOrder:
            _dCharOrder[cPrevious] = {}
        _dCharOrder[cPrevious][cChar] = _dCharOrder[cPrevious].get(cChar, 0) + 1
        cPrevious = cChar


def getCharOrderAfterChar (cChar):
    "return a dictionary of chars with number of times it appears after character <cChar>"
    return _dCharOrder.get(cChar, None)


def displayCharOrder ():
    "display how many times each character appear after another one"
    for key, value in _dCharOrder.items():
        print("[" + key + "]: ", ", ".join([ c+":"+str(n)  for c, n  in  sorted(value.items(), key=lambda t: t[1], reverse=True) ]))



"""
Grammalecte - Compiled regular expressions
"""

import re

#### Lemme
Lemma = re.compile(r"^>(\w[\w-]*)")

#### Analyses
GenderNumber = re.compile(":[mfe]:[spi]")

#### Nom et adjectif
NA = re.compile(":[NA]")

## nombre
NAs = re.compile(":[NA].*:s")
NAp = re.compile(":[NA].*:p")
NAi = re.compile(":[NA].*:i")
NAsi = re.compile(":[NA].*:[si]")
NApi = re.compile(":[NA].*:[pi]")

## genre
NAm = re.compile(":[NA].*:m")
NAf = re.compile(":[NA].*:f")
NAe = re.compile(":[NA].*:e")
NAme = re.compile(":[NA].*:[me]")
NAfe = re.compile(":[NA].*:[fe]")

## nombre et genre
# singuilier
NAms = re.compile(":[NA].*:m.*:s")
NAfs = re.compile(":[NA].*:f.*:s")
NAes = re.compile(":[NA].*:e.*:s")
NAmes = re.compile(":[NA].*:[me].*:s")
NAfes = re.compile(":[NA].*:[fe].*:s")

# singulier et invariable
NAmsi = re.compile(":[NA].*:m.*:[si]")
NAfsi = re.compile(":[NA].*:f.*:[si]")
NAesi = re.compile(":[NA].*:e.*:[si]")
NAmesi = re.compile(":[NA].*:[me].*:[si]")
NAfesi = re.compile(":[NA].*:[fe].*:[si]")

# pluriel
NAmp = re.compile(":[NA].*:m.*:p")
NAfp = re.compile(":[NA].*:f.*:p")
NAep = re.compile(":[NA].*:e.*:p")
NAmep = re.compile(":[NA].*:[me].*:p")
NAfep = re.compile(":[NA].*:[me].*:p")

# pluriel et invariable
NAmpi = re.compile(":[NA].*:m.*:[pi]")
NAfpi = re.compile(":[NA].*:f.*:[pi]")
NAepi = re.compile(":[NA].*:e.*:[pi]")
NAmepi = re.compile(":[NA].*:[me].*:[pi]")
NAfepi = re.compile(":[NA].*:[fe].*:[pi]")

# divers
AD = re.compile(":[AB]")

#### Verbe
Vconj = re.compile(":[123][sp]")
Vconj123 = re.compile(":V[123].*:[123][sp]")

#### Nom | Adjectif | Verbe
NVconj = re.compile(":(?:N|[123][sp])")
NAVconj = re.compile(":(?:N|A|[123][sp])")

#### Spécifique
NnotA = re.compile(":N(?!:A)")
PNnotA = re.compile(":(?:N(?!:A)|Q)")

#### Noms propres
NP = re.compile(":(?:M[12P]|T)")
NPm = re.compile(":(?:M[12P]|T):m")
NPf = re.compile(":(?:M[12P]|T):f")
NPe = re.compile(":(?:M[12P]|T):e")


#### FONCTIONS

def getLemmaOfMorph (s):
    "return lemma in morphology <s>"
    return Lemma.search(s).group(1)

def agreement (l1, l2):
    "returns True if agreement in gender and number is possible between morphologies <l1> and <l2>"
    sGender1, sNumber1 = getGenderNumber(l1)
    sGender2, sNumber2 = getGenderNumber(l2)
    if sNumber1 != ":i" and sNumber2 != ":i" and sNumber1 != sNumber2:
        return False
    if sGender1 != ":e" and sGender2 != ":e" and sGender1 != sGender2:
        return False
    return True

def checkConjVerb (lMorph, sReqConj):
    "returns True if <sReqConj> in <lMorph>"
    return any(sReqConj in s  for s in lMorph)

def getGenderNumber (lMorph):
    "returns tuple (gender, number) of word: (':m', ':f', ':e' or empty string) and (':s', ':p', ':i' or empty string)"
    sGender = ""
    sNumber = ""
    for sMorph in lMorph:
        m = GenderNumber.search(sMorph)
        if m:
            sGenderx = m.group(0)[0:2]
            sNumberx = m.group(0)[2:4]
            if not sGender:
                sGender = sGenderx
            elif sGender != sGenderx:
                sGender = ":e"
            if not sNumber:
                sNumber = sNumberx
            elif sNumber != sNumberx:
                sNumber = ":i"
    return sGender, sNumber


# NOTE :  isWhat (lMorph)    returns True   if lMorph contains nothing else than What
#         mbWhat (lMorph)    returns True   if lMorph contains What at least once

## isXXX = it’s certain

def isNom (lMorph):
    "returns True if all morphologies are “nom”"
    return all(":N" in s  for s in lMorph)

def isNomNotAdj (lMorph):
    "returns True if all morphologies are “nom”, but not “adjectif”"
    return all(NnotA.search(s)  for s in lMorph)

def isAdj (lMorph):
    "returns True if all morphologies are “adjectif”"
    return all(":A" in s  for s in lMorph)

def isNomAdj (lMorph):
    "returns True if all morphologies are “nom” or “adjectif”"
    return all(NA.search(s)  for s in lMorph)

def isNomVconj (lMorph):
    "returns True if all morphologies are “nom” or “verbe conjugué”"
    return all(NVconj.search(s)  for s in lMorph)

def isInv (lMorph):
    "returns True if all morphologies are “invariable”"
    return all(":i" in s  for s in lMorph)

def isSg (lMorph):
    "returns True if all morphologies are “singulier”"
    return all(":s" in s  for s in lMorph)

def isPl (lMorph):
    "returns True if all morphologies are “pluriel”"
    return all(":p" in s  for s in lMorph)

def isEpi (lMorph):
    "returns True if all morphologies are “épicène”"
    return all(":e" in s  for s in lMorph)

def isMas (lMorph):
    "returns True if all morphologies are “masculin”"
    return all(":m" in s  for s in lMorph)

def isFem (lMorph):
    "returns True if all morphologies are “féminin”"
    return all(":f" in s  for s in lMorph)


## mbXXX = MAYBE XXX

def mbNom (lMorph):
    "returns True if one morphology is “nom”"
    return any(":N" in s  for s in lMorph)

def mbAdj (lMorph):
    "returns True if one morphology is “adjectif”"
    return any(":A" in s  for s in lMorph)

def mbAdjNb (lMorph):
    "returns True if one morphology is “adjectif” or “nombre”"
    return any(AD.search(s)  for s in lMorph)

def mbNomAdj (lMorph):
    "returns True if one morphology is “nom” or “adjectif”"
    return any(NA.search(s)  for s in lMorph)

def mbNomNotAdj (lMorph):
    "returns True if one morphology is “nom”, but not “adjectif”"
    bResult = False
    for s in lMorph:
        if ":A" in s:
            return False
        if ":N" in s:
            bResult = True
    return bResult

def mbPpasNomNotAdj (lMorph):
    "returns True if one morphology is “nom” or “participe passé”, but not “adjectif”"
    return any(PNnotA.search(s)  for s in lMorph)

def mbVconj (lMorph):
    "returns True if one morphology is “nom” or “verbe conjugué”"
    return any(Vconj.search(s)  for s in lMorph)

def mbVconj123 (lMorph):
    "returns True if one morphology is “nom” or “verbe conjugué” (but not “avoir” or “être”)"
    return any(Vconj123.search(s)  for s in lMorph)

def mbMG (lMorph):
    "returns True if one morphology is “mot grammatical”"
    return any(":G" in s  for s in lMorph)

def mbInv (lMorph):
    "returns True if one morphology is “invariable”"
    return any(":i" in s  for s in lMorph)

def mbSg (lMorph):
    "returns True if one morphology is “singulier”"
    return any(":s" in s  for s in lMorph)

def mbPl (lMorph):
    "returns True if one morphology is “pluriel”"
    return any(":p" in s  for s in lMorph)

def mbEpi (lMorph):
    "returns True if one morphology is “épicène”"
    return any(":e" in s  for s in lMorph)

def mbMas (lMorph):
    "returns True if one morphology is “masculin”"
    return any(":m" in s  for s in lMorph)

def mbFem (lMorph):
    "returns True if one morphology is “féminin”"
    return any(":f" in s  for s in lMorph)

def mbNpr (lMorph):
    "returns True if one morphology is “nom propre” or “titre de civilité”"
    return any(NP.search(s)  for s in lMorph)

def mbNprMasNotFem (lMorph):
    "returns True if one morphology is “nom propre masculin” but not “féminin”"
    if any(NPf.search(s)  for s in lMorph):
        return False
    return any(NPm.search(s)  for s in lMorph)



"""
Grammalecte - Suggestion phonétique
"""

# License: GPL 3

import re

from .phonet_data import dWord as _dWord
from .phonet_data import lSet as _lSet
from .phonet_data import dMorph as _dMorph


def hasSimil (sWord, sPattern=None):
    "return True if there is list of words phonetically similar to <sWord>"
    if not sWord:
        return False
    if sWord in _dWord:
        if sPattern:
            return any(re.search(sPattern, sMorph)  for sSimil in getSimil(sWord)  for sMorph in _dMorph.get(sSimil, []))
        return True
    if sWord[0:1].isupper():
        sWord = sWord.lower()
        if sWord in _dWord:
            if sPattern:
                return any(re.search(sPattern, sMorph)  for sSimil in getSimil(sWord)  for sMorph in _dMorph.get(sSimil, []))
            return True
    return False


def getSimil (sWord):
    "return list of words phonetically similar to <sWord>"
    if not sWord:
        return []
    if sWord in _dWord:
        return _lSet[_dWord[sWord]]
    if sWord[0:1].isupper():
        sWord = sWord.lower()
        if sWord in _dWord:
            return _lSet[_dWord[sWord]]
    return []


def selectSimil (sWord, sPattern):
    "return a list of words phonetically similar to <sWord> and whom POS is matching <sPattern>"
    if not sPattern:
        return getSimil(sWord)
    aSelect = []
    for sSimil in getSimil(sWord):
        for sMorph in _dMorph.get(sSimil, []):
            if re.search(sPattern, sMorph):
                aSelect.append(sSimil)
    return aSelect


def _getSetNumber (sWord):
    "return the set number where <sWord> belongs, else -1"
    if sWord in _dWord:
        return _dWord[sWord]
    if sWord[0:1].isupper():
        if sWord.lower() in _dWord:
            return _dWord[sWord.lower()]
        if sWord.isupper() and sWord.capitalize() in _dWord:
            return _dWord[sWord.capitalize()]
    return -1


def isSimilAs (sWord, sSimil):
    "return True if <sWord> phonetically similar to <sSimil> (<sWord> tested with several casing)"
    if not sWord or not sSimil:
        return False
    n = _getSetNumber(sWord)
    if n == -1:
        return False
    return n == _getSetNumber(sSimil)



#!python3

"""
INDEXABLE BINARY DIRECT ACYCLIC WORD GRAPH
Implementation of a spellchecker as a transducer (storing transformation code to get lemma and morphologies)
and a spell suggestion mechanism
"""

import traceback
import pkgutil
import re
from functools import wraps
import time
import json
import binascii
import importlib
from collections import OrderedDict
from math import floor

#import logging
#logging.basicConfig(filename="suggestions.log", level=logging.DEBUG)

from . import str_transform as st
from . import char_player as cp
from .echo import echo


def timethis (func):
    "decorator for the execution time"
    @wraps(func)
    def wrapper (*args, **kwargs):
        "something to prevent pylint whining"
        fStart = time.time()
        result = func(*args, **kwargs)
        fEnd = time.time()
        print(func.__name__, fEnd - fStart)
        return result
    return wrapper


class SuggResult:
    """Structure for storing, classifying and filtering suggestions"""

    def __init__ (self, sWord, nSuggLimit=10, nDistLimit=-1):
        self.sWord = sWord
        self.sSimplifiedWord = st.simplifyWord(sWord)
        self.nDistLimit = nDistLimit  if nDistLimit >= 0  else  (len(sWord) // 3) + 1
        self.nMinDist = 1000
        # Temporary sets
        self.aAllSugg = set()   # All suggestions, even the one rejected
        self.dGoodSugg = {}     # Acceptable suggestions
        self.dBestSugg = {}     # Best suggestions
        # Parameters
        self.nSuggLimit = nSuggLimit
        self.nSuggLimitExt = nSuggLimit + 2             # we add few entries in case suggestions merge after casing modifications
        self.nBestSuggLimit = floor(nSuggLimit * 2)     # n times the requested limit
        self.nGoodSuggLimit = nSuggLimit * 15           # n times the requested limit

    def addSugg (self, sSugg, nDeep=0):
        "add a suggestion"
        if sSugg in self.aAllSugg:
            return
        self.aAllSugg.add(sSugg)
        nDistJaro = 1 - st.distanceJaroWinkler(self.sSimplifiedWord, st.simplifyWord(sSugg))
        nDist = floor(nDistJaro * 10)
        if nDist < self.nMinDist:
            self.nMinDist = nDist
        #logging.info((nDeep * "  ") + "__" + sSugg + "__ " + str(round(nDistJaro*1000)))
        if nDistJaro < .11:     # Best suggestions
            self.dBestSugg[sSugg] = round(nDistJaro*1000)
            if len(self.dBestSugg) > self.nBestSuggLimit:
                self.nDistLimit = -1  # make suggest() to end search
        elif nDistJaro < .33:   # Good suggestions
            self.dGoodSugg[sSugg] = round(nDistJaro*1000)
            if len(self.dGoodSugg) > self.nGoodSuggLimit:
                self.nDistLimit = -1  # make suggest() to end search
        self.nDistLimit = min(self.nDistLimit, self.nMinDist+1)

    def getSuggestions (self):
        "return a list of suggestions"
        # we sort the better results with the original word
        lRes = []
        if len(self.dBestSugg) > 0:
            # sort only with simplified words
            lResTmp = sorted(self.dBestSugg.items(), key=lambda x: x[1])
            for i in range(min(self.nSuggLimitExt, len(lResTmp))):
                lRes.append(lResTmp[i][0])
        if len(lRes) < self.nSuggLimitExt:
            # sort with simplified words and original word
            lResTmp = sorted(self.dGoodSugg.items(), key=lambda x: ((1-st.distanceJaroWinkler(self.sWord, x[0]))*10, x[1]))
            for i in range(min(self.nSuggLimitExt, len(lResTmp))):
                lRes.append(lResTmp[i][0])
        # casing
        if self.sWord.isupper():
            lRes = list(OrderedDict.fromkeys(map(lambda sSugg: sSugg.upper(), lRes))) # use dict, when Python 3.6+
        elif self.sWord[0:1].isupper():
            # dont’ use <.istitle>
            lRes = list(OrderedDict.fromkeys(map(lambda sSugg: sSugg[0:1].upper()+sSugg[1:], lRes))) # use dict, when Python 3.6+
        return lRes[:self.nSuggLimit]


class IBDAWG:
    """INDEXABLE BINARY DIRECT ACYCLIC WORD GRAPH"""

    def __init__ (self, source):
        if isinstance(source, str):
            by = pkgutil.get_data(__package__, "_dictionaries/" + source)
            if not by:
                raise OSError("# Error. File not found or not loadable: "+source)
            self.sFileName = source
            oData = json.loads(by.decode("utf-8"))     #json.loads(by)    # In Python 3.6, can read directly binary strings
        else:
            self.sFileName = "[None]"
            oData = source

        self.__dict__.update(oData)
        self.dCharVal = { v: k  for k, v in self.dChar.items() }
        self.a2grams = set(getattr(self, 'l2grams'))  if hasattr(self, 'l2grams')  else None

        if "lByDic" not in oData:
            print(">>>> lByDic not in oData")
            if "sByDic" not in oData:
                raise TypeError("# Error. No usable data in the dictionary.")
            # old dictionary version
            self.lByDic = []
            self.byDic = binascii.unhexlify(oData["sByDic"])
            nAcc = 0
            byBuffer = b""
            nDivisor = (self.nBytesArc + self.nBytesNodeAddress) / 2
            for i in range(0, len(self.byDic)):
                byBuffer += self.byDic[i:i+1]
                if nAcc == (self.nBytesArc - 1):
                    self.lByDic.append(int.from_bytes(byBuffer, byteorder="big"))
                    byBuffer = b""
                elif nAcc == (self.nBytesArc + self.nBytesNodeAddress - 1):
                    self.lByDic.append(round(int.from_bytes(byBuffer, byteorder="big") / nDivisor))
                    byBuffer = b""
                    nAcc = -1
                nAcc = nAcc + 1

        # masks
        self._arcMask = (2 ** ((self.nBytesArc * 8) - 3)) - 1
        self._finalNodeMask = 1 << ((self.nBytesArc * 8) - 1)
        self._lastArcMask = 1 << ((self.nBytesArc * 8) - 2)

        # function to decode the affix/suffix code
        if self.cStemming == "S":
            self.funcStemming = st.changeWordWithSuffixCode
        elif self.cStemming == "A":
            self.funcStemming = st.changeWordWithAffixCode
        else:
            self.funcStemming = st.noStemming

        self.bAcronymValid = False
        self.bNumAtLastValid = False

        # lexicographer module ?
        self.lexicographer = None
        try:
            self.lexicographer = importlib.import_module(".lexgraph_"+self.sLangCode, "grammalecte.graphspell")
        except ImportError:
            print("# No module <graphspell.lexgraph_"+self.sLangCode+".py>")

    def getInfo (self):
        "return string about the IBDAWG"
        return  "  Language: {0.sLangName}   Lang code: {0.sLangCode}   Dictionary name: {0.sDicName}" \
                "  Date: {0.sDate}   Stemming: {0.cStemming}FX\n" \
                "  Arcs values:  {0.nArcVal:>10,} = {0.nChar:>5,} characters,  {0.nAff:>6,} affixes,  {0.nTag:>6,} tags\n" \
                "  Dictionary: {0.nEntry:>12,} entries,    {0.nNode:>11,} nodes,   {0.nArc:>11,} arcs\n" \
                "  Address size: {0.nBytesNodeAddress:>1} bytes,  Arc size: {0.nBytesArc:>1} bytes\n".format(self)

    def isValidToken (self, sToken):
        "checks if <sToken> is valid (if there is hyphens in <sToken>, <sToken> is split, each part is checked)"
        sToken = st.spellingNormalization(sToken)
        if self.isValid(sToken):
            return True
        if "-" in sToken:
            if sToken.count("-") > 4:
                return True
            return all(self.isValid(sWord)  for sWord in sToken.split("-"))
        if "." in sToken or "·" in sToken:
            return True
        return False

    def isValid (self, sWord):
        "checks if <sWord> is valid (different casing tested if the first letter is a capital)"
        if not sWord:
            return True
        if self.lookup(sWord):
            return True
        if sWord[0:1].isupper():
            if len(sWord) > 1:
                if sWord.istitle():
                    return self.lookup(sWord.lower())
                if sWord.isupper():
                    return self.bAcronymValid or self.lookup(sWord.lower()) or self.lookup(sWord.capitalize())
                return self.lookup(sWord[:1].lower() + sWord[1:])
            return self.lookup(sWord.lower())
        if sWord[0:1].isdigit():
            return True
        return False

    def lookup (self, sWord):
        "returns True if <sWord> in dictionary (strict verification)"
        iAddr = 0
        for c in sWord:
            if c not in self.dChar:
                return False
            iAddr = self._lookupArcNode(self.dChar[c], iAddr)
            if iAddr is None:
                return False
        return bool(self.lByDic[iAddr] & self._finalNodeMask)

    def getMorph (self, sWord):
        "retrieves morphologies list, different casing allowed"
        if not sWord:
            return []
        sWord = st.spellingNormalization(sWord)
        l = self._morph(sWord)
        if sWord[0:1].isupper():
            l.extend(self._morph(sWord.lower()))
            if sWord.isupper() and len(sWord) > 1:
                l.extend(self._morph(sWord.capitalize()))
        return l

    #@timethis
    def suggest (self, sWord, nSuggLimit=10, bSplitTrailingNumbers=False):
        "returns a set of suggestions for <sWord>"
        sWord = sWord.rstrip(".")   # useful for LibreOffice
        sWord = st.spellingNormalization(sWord)
        sPfx = ""
        sSfx = ""
        if self.lexicographer:
            sPfx, sWord, sSfx = self.lexicographer.split(sWord)
        nMaxSwitch = max(len(sWord) // 3, 1)
        nMaxDel = len(sWord) // 5
        nMaxHardRepl = max((len(sWord) - 5) // 4, 1)
        nMaxJump = max(len(sWord) // 4, 1)
        oSuggResult = SuggResult(sWord, nSuggLimit)
        sWord = st.cleanWord(sWord)
        if bSplitTrailingNumbers:
            self._splitTrailingNumbers(oSuggResult, sWord)
        self._splitSuggest(oSuggResult, sWord)
        self._suggest(oSuggResult, sWord)
        self._suggest(oSuggResult, sWord, nMaxSwitch, nMaxDel, nMaxHardRepl, nMaxJump)
        aSugg = oSuggResult.getSuggestions()
        if sSfx or sPfx:
            # we add what we removed
            return list(map(lambda sSug: sPfx + sSug + sSfx, aSugg))
        return aSugg

    def _splitTrailingNumbers (self, oSuggResult, sWord):
        m = re.match(r"(\D+)([0-9]+)$", sWord)
        if m and m.group(1)[-1:].isalpha():
            oSuggResult.addSugg(m.group(1) + " " + st.numbersToExponent(m.group(2)))

    def _splitSuggest (self, oSuggResult, sWord):
        # split at apostrophes
        for cSplitter in "'’":
            if cSplitter in sWord:
                sWord1, sWord2 = sWord.split(cSplitter, 1)
                if self.isValid(sWord1) and self.isValid(sWord2):
                    oSuggResult.addSugg(sWord1+" "+sWord2)

    def _suggest (self, oSuggResult, sRemain, nMaxSwitch=0, nMaxDel=0, nMaxHardRepl=0, nMaxJump=0, nDist=0, nDeep=0, iAddr=0, sNewWord="", bAvoidLoop=False):
        # recursive function
        #logging.info((nDeep * "  ") + sNewWord + ":" + sRemain)
        if self.lByDic[iAddr] & self._finalNodeMask:
            if not sRemain:
                oSuggResult.addSugg(sNewWord, nDeep)
                for sTail in self._getTails(iAddr):
                    oSuggResult.addSugg(sNewWord+sTail, nDeep)
                return
            if (len(sNewWord) + len(sRemain) == len(oSuggResult.sWord)) and oSuggResult.sWord.lower().startswith(sNewWord.lower()) and self.isValid(sRemain):
                if self.sLangCode == "fr" and sNewWord.lower() in ("l", "d", "n", "m", "t", "s", "c", "j", "qu", "lorsqu", "puisqu", "quoiqu", "jusqu", "quelqu") and sRemain[0:1] in cp.aVowel:
                    oSuggResult.addSugg(sNewWord+"’"+sRemain, nDeep)
                if (len(sNewWord) > 1 and len(sRemain) > 1) or sNewWord in "aày" or sRemain in "aày":
                    oSuggResult.addSugg(sNewWord+" "+sRemain, nDeep)
        if nDist > oSuggResult.nDistLimit:
            return
        cCurrent = sRemain[0:1]
        for cChar, jAddr in self._getCharArcs(iAddr):
            if cChar in cp.d1to1.get(cCurrent, cCurrent):
                self._suggest(oSuggResult, sRemain[1:], nMaxSwitch, nMaxDel, nMaxHardRepl, nMaxJump, nDist, nDeep+1, jAddr, sNewWord+cChar)
            elif not bAvoidLoop:
                if nMaxHardRepl and self.isNgramsOK(cChar+sRemain[1:2]):
                    self._suggest(oSuggResult, sRemain[1:], nMaxSwitch, nMaxDel, nMaxHardRepl-1, nMaxJump, nDist+1, nDeep+1, jAddr, sNewWord+cChar, True)
                if nMaxJump:
                    self._suggest(oSuggResult, sRemain, nMaxSwitch, nMaxDel, nMaxHardRepl, nMaxJump-1, nDist+1, nDeep+1, jAddr, sNewWord+cChar, True) # True for avoiding loop?
        if not bAvoidLoop: # avoid infinite loop
            if len(sRemain) > 1:
                if cCurrent == sRemain[1:2]:
                    # same char, we remove 1 char without adding 1 to <sNewWord>
                    self._suggest(oSuggResult, sRemain[1:], nMaxSwitch, nMaxDel, nMaxHardRepl, nMaxJump, nDist, nDeep+1, iAddr, sNewWord)
                else:
                    # switching chars
                    if nMaxSwitch and self.isNgramsOK(sNewWord[-1:]+sRemain[1:2]) and self.isNgramsOK(sRemain[1:2]+sRemain[0:1]):
                        self._suggest(oSuggResult, sRemain[1:2]+sRemain[0:1]+sRemain[2:], nMaxSwitch-1, nMaxDel, nMaxHardRepl, nMaxJump, nDist+1, nDeep+1, iAddr, sNewWord, True)
                    # delete char
                    if nMaxDel and self.isNgramsOK(sNewWord[-1:]+sRemain[1:2]):
                        self._suggest(oSuggResult, sRemain[1:], nMaxSwitch, nMaxDel-1, nMaxHardRepl, nMaxJump, nDist+1, nDeep+1, iAddr, sNewWord, True)
                # Phonetic replacements
                for sRepl in cp.get1toXReplacement(sNewWord[-1:], cCurrent, sRemain[1:2]):
                    self._suggest(oSuggResult, sRepl + sRemain[1:], nMaxSwitch, nMaxDel, nMaxHardRepl, nMaxJump, nDist, nDeep+1, iAddr, sNewWord, True)
                for sRepl in cp.d2toX.get(sRemain[0:2], ()):
                    self._suggest(oSuggResult, sRepl + sRemain[2:], nMaxSwitch, nMaxDel, nMaxHardRepl, nMaxJump, nDist, nDeep+1, iAddr, sNewWord, True)
            # end of word
            if len(sRemain) == 2:
                for sRepl in cp.dFinal2.get(sRemain, ()):
                    self._suggest(oSuggResult, sRepl, nMaxSwitch, nMaxDel, nMaxHardRepl, nMaxJump, nDist, nDeep+1, iAddr, sNewWord, True)
            elif len(sRemain) == 1:
                self._suggest(oSuggResult, "", nMaxSwitch, nMaxDel, nMaxHardRepl, nMaxJump, nDist, nDeep+1, iAddr, sNewWord, True) # remove last char and go on
                for sRepl in cp.dFinal1.get(sRemain, ()):
                    self._suggest(oSuggResult, sRepl, nMaxSwitch, nMaxDel, nMaxHardRepl, nMaxJump, nDist, nDeep+1, iAddr, sNewWord, True)

    def isNgramsOK (self, sChars):
        "returns True if sChars in known 2grams"
        if len(sChars) != 2:
            return True
        if not self.a2grams:
            return True
        return sChars in self.a2grams

    def _getCharArcs (self, iAddr):
        "generator: yield all chars and addresses from node at address <iAddr>"
        for nVal, jAddr in self._getArcs(iAddr):
            if nVal <= self.nChar:
                yield (self.dCharVal[nVal], jAddr)

    def _getTails (self, iAddr, sTail="", n=2):
        "return a list of suffixes ending at a distance of <n> from <iAddr>"
        aTails = set()
        for nVal, jAddr in self._getArcs(iAddr):
            if nVal <= self.nChar:
                if self.lByDic[jAddr] & self._finalNodeMask:
                    aTails.add(sTail + self.dCharVal[nVal])
                if n and not aTails:
                    aTails.update(self._getTails(jAddr, sTail+self.dCharVal[nVal], n-1))
        return aTails

    def drawPath (self, sWord, iAddr=0):
        "show the path taken by <sWord> in the graph"
        sWord = st.spellingNormalization(sWord)
        c1 = sWord[0:1]  if sWord  else " "
        iPos = -1
        n = 0
        echo(c1 + ": ", end="")
        for c2, jAddr in self._getCharArcs(iAddr):
            echo(c2, end="")
            if c2 == sWord[0:1]:
                iNextNodeAddr = jAddr
                iPos = n
            n += 1
        if not sWord:
            return
        if iPos >= 0:
            echo("\n   " + " " * iPos + "|")
            self.drawPath(sWord[1:], iNextNodeAddr)

    def getSimilarEntries (self, sWord, nSuggLimit=10):
        "return a list of tuples (similar word, stem, morphology)"
        if not sWord:
            return []
        lResult = []
        for sSimilar in self.suggest(sWord, nSuggLimit):
            for sMorph in self.getMorph(sSimilar):
                nCut = sMorph.find("/")
                lResult.append( (sSimilar, sMorph[1:nCut], sMorph[nCut+1:]) )
        return lResult

    def select (self, sFlexPattern="", sTagsPattern=""):
        "generator: returns all entries which flexion fits <sFlexPattern> and morphology fits <sTagsPattern>"
        zFlexPattern = None
        zTagsPattern = None
        try:
            if sFlexPattern:
                zFlexPattern = re.compile(sFlexPattern)
            if sTagsPattern:
                zTagsPattern = re.compile(sTagsPattern)
        except re.error:
            print("# Error in regex pattern")
            traceback.print_exc()
        yield from self._select(zFlexPattern, zTagsPattern, 0, "")

    def _select (self, zFlexPattern, zTagsPattern, iAddr, sWord):
        # recursive generator
        for nVal, jAddr in self._getArcs(iAddr):
            if nVal <= self.nChar:
                # simple character
                yield from self._select(zFlexPattern, zTagsPattern, jAddr, sWord + self.lArcVal[nVal])
            else:
                if not zFlexPattern or zFlexPattern.search(sWord):
                    sStem = self.funcStemming(sWord, self.lArcVal[nVal])
                    for nMorphVal, _ in self._getArcs(jAddr):
                        if not zTagsPattern or zTagsPattern.search(self.lArcVal[nMorphVal]):
                            yield [sWord, sStem, self.lArcVal[nMorphVal]]

    def _morph (self, sWord):
        "returns morphologies of <sWord>"
        iAddr = 0
        for c in sWord:
            if c not in self.dChar:
                return []
            iAddr = self._lookupArcNode(self.dChar[c], iAddr)
            if iAddr is None:
                return []
        if self.lByDic[iAddr] & self._finalNodeMask:
            l = []
            nRawArc = 0
            while not nRawArc & self._lastArcMask:
                iEndArcAddr = iAddr + 1
                nRawArc = self.lByDic[iAddr]
                nArc = nRawArc & self._arcMask
                if nArc > self.nChar:
                    # This value is not a char, this is a stemming code
                    sStem = ">" + self.funcStemming(sWord, self.lArcVal[nArc])
                    # Now , we go to the next node and retrieve all following arcs values, all of them are tags
                    iAddr2 = self.lByDic[iEndArcAddr]
                    nRawArc2 = 0
                    while not nRawArc2 & self._lastArcMask:
                        iEndArcAddr2 = iAddr2 + 1
                        nRawArc2 = self.lByDic[iAddr2]
                        l.append(sStem + "/" + self.lArcVal[nRawArc2 & self._arcMask])
                        iAddr2 = iEndArcAddr2 + 1
                iAddr = iEndArcAddr + 1
            return l
        return []

    def _stem (self, sWord):
        "returns stems list of <sWord>"
        iAddr = 0
        for c in sWord:
            if c not in self.dChar:
                return []
            iAddr = self._lookupArcNode(self.dChar[c], iAddr)
            if iAddr is None:
                return []
        if self.lByDic[iAddr] & self._finalNodeMask:
            l = []
            nRawArc = 0
            while not nRawArc & self._lastArcMask:
                iEndArcAddr = iAddr + 1
                nRawArc = self.lByDic[iAddr]
                nArc = nRawArc & self._arcMask
                if nArc > self.nChar:
                    # This value is not a char, this is a stemming code
                    l.append(self.funcStemming(sWord, self.lArcVal[nArc]))
                iAddr = iEndArcAddr + 1
            return l
        return []

    def _lookupArcNode (self, nVal, iAddr):
        "looks if <nVal> is an arc at the node at <iAddr>, if yes, returns address of next node else None"
        while True:
            iEndArcAddr = iAddr + 1
            nRawArc = self.lByDic[iAddr]
            if nVal == (nRawArc & self._arcMask):
                # the value we are looking for
                # we return the address of the next node
                return self.lByDic[iEndArcAddr]
            # value not found
            if nRawArc & self._lastArcMask:
                return None
            iAddr = iEndArcAddr + 1

    def _getArcs (self, iAddr):
        "generator: return all arcs at <iAddr> as tuples of (nVal, iAddr)"
        while True:
            iEndArcAddr = iAddr + 1
            nRawArc = self.lByDic[iAddr]
            yield nRawArc & self._arcMask, self.lByDic[iEndArcAddr]
            if nRawArc & self._lastArcMask:
                break
            iAddr = iEndArcAddr + 1

    def _writeNodes (self, spfDest):
        "for debugging only"
        print(" > Write binary nodes")
        with open(spfDest, 'w', 'utf-8', newline="\n") as hDst:
            iAddr = 0
            hDst.write("i{:_>10} -- #{:_>10}\n".format("0", iAddr))
            while iAddr < len(self.lByDic):
                iEndArcAddr = iAddr + 1
                nRawArc = self.lByDic[iAddr]
                nArc = nRawArc & self._arcMask
                hDst.write("  {:<20}  {:0>16}  i{:>10}   #{:_>10}\n".format(self.lArcVal[nArc], bin(nRawArc)[2:], "?", self.lByDic[iEndArcAddr]))
                iAddr = iEndArcAddr + 1
                if (nRawArc & self._lastArcMask) and iAddr < len(self.lByDic):
                    hDst.write("\ni{:_>10} -- #{:_>10}\n".format("?", iAddr))
            hDst.close()



"""
Grammalecte, grammar checker
"""

import importlib
import json

from . import text


class GrammarChecker:
    "GrammarChecker: Wrapper for the grammar checker engine"

    def __init__ (self, sLangCode, sContext="Python"):
        self.sLangCode = sLangCode
        # Grammar checker engine
        self.gce = importlib.import_module("."+sLangCode, "grammalecte")
        self.gce.load(sContext)
        # Spell checker
        self.oSpellChecker = self.gce.getSpellChecker()
        # Lexicographer
        self.oLexicographer = None
        # Text formatter
        self.oTextFormatter = None

    def getGCEngine (self):
        "return the grammar checker object"
        return self.gce

    def getSpellChecker (self):
        "return the spell checker object"
        return self.oSpellChecker

    def getTextFormatter (self):
        "load and return the text formatter"
        if self.oTextFormatter is None:
            tf = importlib.import_module("."+self.sLangCode+".textformatter", "grammalecte")
            self.oTextFormatter = tf.TextFormatter()
        return self.oTextFormatter

    def getLexicographer (self):
        "load and return the lexicographer"
        if self.oLexicographer is None:
            lxg = importlib.import_module("."+self.sLangCode+".lexicographe", "grammalecte")
            self.oLexicographer = lxg.Lexicographe(self.oSpellChecker)
        return self.oLexicographer

    def displayGCOptions (self):
        "display the grammar checker options"
        self.gce.displayOptions()

    def getParagraphErrors (self, sText, dOptions=None, bContext=False, bSpellSugg=False, bDebug=False):
        "returns a tuple: (grammar errors, spelling errors)"
        aGrammErrs = self.gce.parse(sText, "FR", bDebug=bDebug, dOptions=dOptions, bContext=bContext)
        aSpellErrs = self.oSpellChecker.parseParagraph(sText, bSpellSugg)
        return aGrammErrs, aSpellErrs

    def getParagraphWithErrors (self, sText, dOptions=None, bEmptyIfNoErrors=False, bSpellSugg=False, nWidth=100, bDebug=False):
        "parse text and return a readable text with underline errors"
        aGrammErrs, aSpellErrs = self.getParagraphErrors(sText, dOptions, False, bSpellSugg, bDebug)
        if bEmptyIfNoErrors and not aGrammErrs and not aSpellErrs:
            return ("", [])
        return text.generateParagraph(sText, aGrammErrs, aSpellErrs, nWidth)

    def getTextWithErrors (self, sText, bEmptyIfNoErrors=False, bSpellSugg=False, nWidth=100, bDebug=False):
        "[todo]"

    def getParagraphErrorsAsJSON (self, iIndex, sText, dOptions=None, bContext=False, bEmptyIfNoErrors=False, bSpellSugg=False, bReturnText=False, lLineSet=None, bDebug=False):
        "parse text and return errors as a JSON string"
        aGrammErrs, aSpellErrs = self.getParagraphErrors(sText, dOptions, bContext, bSpellSugg, bDebug)
        aGrammErrs = list(aGrammErrs)
        if bEmptyIfNoErrors and not aGrammErrs and not aSpellErrs:
            return ""
        if lLineSet:
            aGrammErrs, aSpellErrs = text.convertToXY(aGrammErrs, aSpellErrs, lLineSet)
            return json.dumps({ "lGrammarErrors": aGrammErrs, "lSpellingErrors": aSpellErrs }, ensure_ascii=False)
        if bReturnText:
            return json.dumps({ "iParagraph": iIndex, "sText": sText, "lGrammarErrors": aGrammErrs, "lSpellingErrors": aSpellErrs }, ensure_ascii=False)
        return json.dumps({ "iParagraph": iIndex, "lGrammarErrors": aGrammErrs, "lSpellingErrors": aSpellErrs }, ensure_ascii=False)

    def getTextErrorsAsJSON (self, sText, bContext=False, bEmptyIfNoErrors=False, bSpellSugg=False, bReturnText=False, bDebug=False):
        "[todo]"



"""
Grammar checker regex rules
"""

# generated code, do not edit

lParagraphRules = [
  ["tab", [
    ['^[ \xa0\u202f   ]+', False, '#357p', 'tab_début_ligne', 4, [[None, '-', '', 0, 'Espace·s en début de ligne à supprimer\xa0: utilisez les retraits de paragraphe.', '']]],
    ['[ \xa0\u202f    ]+$', False, '#358p', 'tab_fin_ligne', 4, [[None, '-', '', 0, 'Espace·s en fin de ligne à supprimer.', '']]],
  ]],
  ["esp", [
    ['^[ \xa0\u202f]+', False, '#366p', 'esp_début_ligne', 4, [[None, '-', '', 0, 'Espace·s en début de ligne à supprimer\xa0: utilisez les retraits de paragraphe (ou les tabulations à la rigueur).', '']]],
    ['[ \xa0\u202f]+$', False, '#367p', 'esp_fin_ligne', 4, [[None, '-', '', 0, 'Espace·s en fin de ligne à supprimer.', '']]],
    ['(?:\\b|[.?!,:;%‰‱˚»”])(  +)', False, '#368p', 'esp_milieu_ligne', 4, [[None, '-', ' ', 1, 'Espace·s surnuméraire·s à supprimer.', '']]],
    ['(?: \xa0|\xa0 )', False, '#378p', 'esp_mélangés1', 4, [[None, '-', ' |\xa0', 0, 'Incohérence\xa0: l’espace insécable à côté de l’espace sécable n’a pas d’effet.', '']]],
    ['(?:\\u0009[ \xa0]|[ \xa0]\\u0009)', False, '#379p', 'esp_mélangés2', 4, [[None, '-', ' |\t', 0, 'Incohérence\xa0: espace et tabulation.', '']]],
    ['[\xa0\u202f][\xa0\u202f]+', False, '#386p', 'esp_insécables_multiples', 4, [[None, '-', ' |\u202f', 0, 'Espace·s insécable·s surnuméraire·s à supprimer.', '']]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]*)( [-–—]|[-–—] )(\\w[\\w-]*)(?![\\w–-])', True, '#393p', 'esp_avant_après_tiret', 4, [['_c_esp_avant_après_tiret_1', '>', ''], ['_c_esp_avant_après_tiret_2', '-', '\\1-\\3|\\1 – \\3|\\1 — \\3', 0, 'Espace superflu s’il s’agit bien d’une forme conjuguée interrogative (sinon, il manque un espace à côté du tiret).', ''], ['_c_esp_avant_après_tiret_3', '-', ' – | — ', 2, 'Il manque un espace à côté du tiret.', '']]],
  ]],
  ["typo", [
    ['([\\[(]) ', False, '#416p', 'typo_espace_après_signe_ouvrant', 4, [[None, '-', '\\1', 0, 'Pas d’espace après ce signe.', '']]],
    [' ([\\])])', False, '#417p', 'typo_espace_avant_signe_fermant', 4, [[None, '-', '\\1', 0, 'Pas d’espace avant ce signe.', '']]],
    ['[)]\\b(?![s¹²³⁴⁵⁶⁷⁸⁹⁰]\\b)', False, '#427p', 'typo_parenthèse_fermante_collée', 4, [['_c_typo_parenthèse_fermante_collée_1', '-', ') ', 0, 'Il manque un espace après la parenthèse.', '']]],
    ['\\b[(](?=[^)][^)][^)])', False, '#430p', 'typo_parenthèse_ouvrante_collée', 4, [[None, '-', ' (', 0, 'Il manque un espace avant la parenthèse.', '']]],
    ['[ \xa0][.](?=[ \xa0])', False, '#441p', 'typo_point_entre_deux_espaces', 4, [[None, '-', '.', 0, 'Pas d’espace avant un point.', '']]],
    ['[ \xa0][.](?=\\w)', False, '#442p', 'typo_point_collé_à_mot_suivant', 4, [[None, '-', '. |.', 0, 'L’espace se place après le point, et non avant (ou bien sert de liant entre deux lettres).', '']]],
    ['\\w+([ \xa0]…)', False, '#449p', 'typo_points_suspension_après_espace', 4, [[None, '-', '…', 1, 'Pas d’espace avant le signe “…”.', '']]],
    ['[ \xa0],(?=[ \xa0])', False, '#457p', 'typo_virgule_entre_deux_espaces', 4, [[None, '-', ',', 0, 'Pas d’espace avant une virgule.', '']]],
    ['[ \xa0],(?=\\w)', False, '#458p', 'typo_virgule_collée_à_mot_suivant', 4, [[None, '-', ', ', 0, 'Pas d’espace avant une virgule. Un espace après.', '']]],
    ['[ \xa0]”[ \xa0]', False, '#465p', 'typo_guillemets_fermants_perdus', 4, [[None, '-', '” | “', 0, 'Pas d’espace avant ces guillemets ouvrants.', '']]],
    ['[ \xa0]“[ \xa0]', False, '#468p', 'typo_guillemets_ouvrants_perdus', 4, [[None, '-', ' “|” ', 0, 'Pas d’espace après ces guillemets ouvrants.', '']]],
    ['[ \xa0\u202f](?:"|[\'‘’][\'‘’])[ \xa0\u202f]', False, '#471p', 'typo_guillemets_perdus', 4, [[None, '-', ' «\xa0|\xa0» | “|” ', 0, 'Guillemets isolés.', '']]],
    ['^(?:"|[\'‘’][\'‘’])[ \xa0\u202f]', False, '#474p', 'typo_commencement_guillemets', 4, [[None, '-', '«\xa0|“', 0, 'Guillemets ouvrants.', '']]],
    ['([ \xa0\u202f](?:"|[\'‘’][\'‘’]))(?:$|[.,;?!])', False, '#477p', 'typo_guillemets_fin', 4, [[None, '-', '\xa0»|”', 1, 'Guillemets fermants.', '']]],
    ['[ \xa0]”(?=\\w)', False, '#488p', 'typo_guillemets_fermants_inverses', 4, [[None, '-', ' “', 0, 'Mettez des guillemets ouvrants plutôt que fermants.', '']]],
    ['\\w([«“][ \xa0])', False, '#491p', 'typo_guillemets_ouvrants_inverses1', 4, [[None, '-', '\xa0»|”', 1, 'Mettez des guillemets fermants plutôt qu’ouvrants.', '']]],
    ['([«“])[ \xa0]*$', False, '#494p', 'typo_guillemets_ouvrants_inverses2', 4, [[None, '-', '\xa0»|”', 1, 'Mettez des guillemets fermants plutôt qu’ouvrants.', '']]],
    ['\\w([ \xa0][«“])[ \xa0]?[!?,:;.…](?![ \xa0]?[»”])', False, '#497p', 'typo_guillemets_ouvrants_inverses3', 4, [[None, '-', '\xa0»|”', 1, 'Mettez des guillemets fermants plutôt qu’ouvrants.', '']]],
  ]],
  ["", [
    ['(?i)\\w[\\w.-]*@\\w[\\w.-]*\\w[.]\\w\\w+', True, '#524p', 'p_email', 4, [[None, '~', '__MAIL__', 0]]],
    ['(?i)(?:ht|f)tps?://[\\w./?&!%=+*"\'@$#-]+', True, '#528p', 'p_URL', 4, [[None, '~', '__URL__', 0]]],
    ['(?i)((?:\\w[\\w-]*[.])*)(\\w[\\w-]+)([.](?:com|net|org|info|fr|ca|be|ch|i[ot]|co[.]uk|tk|jp|zh|ru|us|nl|xyz))(?![\\w–-])', True, '#530p', 'p_URL2', 4, [[None, '~', '*', 1], [None, '~', '=_p_p_URL2_2', 2], [None, '~', '*', 3]]],
    ['(?i)^\\d+[.][\\d.-]*', True, '#537p', 'p_chapitre', 4, [[None, '~', '*', 0]]],
    ['(?i)^\\d+(?:  +|\\t+)', True, '#541p', 'p_num_chapitre', 4, [[None, '~', '*', 0]]],
    ['(?<![\\w.,–-])[cC](?:f|hap|oll?)([.])', False, '#545p', 'p_chap_coll_cf', 4, [[None, '~', '*', 1]]],
    ['(?<![\\w.,–-])p[.] ?\\d+(?:-\\d+|)(?![\\w–-])', False, '#546p', 'p_page', 4, [[None, '~', '*', 0]]],
    ['(?<![\\w.,–-])pp[.] ?\\d+-\\d+(?![\\w–-])', False, '#547p', 'p_pages', 4, [[None, '~', '*', 0]]],
    ['(?i)(?<![\\w.,–-])figure (\\d+(?:[.:-]\\d+|))(?![\\w–-])', True, '#548p', 'p_figure', 4, [[None, '~', '*', 1]]],
    ['(?i)(?<![\\w.,–-])art(?:icle|[.]) (1(?:er|ᵉʳ)|\\d+(?:[.:-]\\w+|))(?![\\w–-])', True, '#549p', 'p_article', 4, [[None, '~', '*', 1]]],
    ['(?<![\\w.,–-])av[.]', False, '#550p', 'p_av', 4, [[None, '~', 'av', 0]]],
    ['(?<![\\w.,–-])vs[.]', False, '#553p', 'p_vs', 4, [[None, '-', 'vs ', 0, 'Pas de point après cette abréviation.', ''], [None, '~', 'vs', 0]]],
    ['(?i)(?<![\\w.,–-])(?:versions?|mises? [àa] jour) ([0-9]+(?:[.][0-9]+[a-z]?)*)(?![\\w–-])', True, '#559p', 'p_version_numéro', 4, [[None, '~', '*', 1]]],
    ['([A-Z][.][A-Z][.](?:[A-Z][.])*) +[A-ZÉÀÂÊÎÈÔ]', False, '#563p', 'p_sigle1', 4, [[None, '~', '=_p_p_sigle1_1', 1]]],
    ['(?<![\\w.,–-])[a-zA-Z][.][a-zA-Z][.](?:[a-zA-Z][.])*', False, '#565p', 'p_sigle2', 4, [['_c_p_sigle2_1', '>', ''], ['_c_p_sigle2_2', '-', '=_s_p_sigle2_2', 0, 'Sigle. Il est recommandé d’ôter les points pour les sigles. (S’il s’agit d’un prénom et d’un nom, mettez un espace.)', 'https://fr.wikipedia.org/wiki/Sigle#Typographie'], ['_c_p_sigle2_3', '-', '=_s_p_sigle2_3', 0, 'Sigle. Il est recommandé d’ôter les points pour les sigles.', 'https://fr.wikipedia.org/wiki/Sigle#Typographie'], ['_c_p_sigle2_4', '~', '=_p_p_sigle2_4', 0]]],
    ['(?<![\\w.,–-])J[.]-[A-Z][.]', False, '#572p', 'p_sigle3', 4, [[None, '~', '=_p_p_sigle3_1', 0]]],
    ["(?<![\\w.,–-])(M[.]) (?:[A-ZÉÈÎ]\\w*|l[ae'’])", False, '#576p', 'p_M_point', 4, [[None, '~', 'Mr', 1]]],
    ['(?<![\\w.,–-])MM[.]', False, '#578p', 'p_MM_point', 4, [[None, '~', 'MM ', 0]]],
    ['(?<![\\w.,–-])M(?:r|gr|me) [A-ZÉ]([.])(?=\\W+[a-zéèêâîïû])', False, '#580p', 'p_Mr_Mgr_Mme_point', 4, [[None, '~', '*', 1]]],
    ['(?<![\\w.,–-])([A-ZÉÈÂÎ][\\w-]+)[ \xa0][A-ZÉÈÂ]([.])[ \xa0]([A-ZÉÈÂ][\\w-]+)(?![\\w–-])', False, '#584p', 'p_prénom_lettre_point_patronyme', 4, [['_c_p_prénom_lettre_point_patronyme_1', '~', '*', 2]]],
    ['(?<![\\w.,–-])([A-ZÉÈÂÎ][\\w-]+)[ \xa0][A-ZÉÈÂ]([.])', False, '#587p', 'p_prénom_lettre_point', 4, [['_c_p_prénom_lettre_point_1', '~', '_', 2]]],
    ['(?<![\\w.,–-])[A-ZÉÈÂÎ][\\w-]+[-–—]L(?:es?|a) [A-ZÉÈÂÎ][\\w-]+(?![\\w–-])', False, '#592p', 'p_patronyme_composé_avec_le_la_les', 4, [[None, '~', '=_p_p_patronyme_composé_avec_le_la_les_1', 0]]],
    ['(?<![\\w.,–-])\\d+[.:]\\d+[.:]\\d+[.:]\\d+(?![\\w–-])', False, '#596p', 'p_adresse_IP', 4, [[None, '~', '*', 0]]],
    ['(?<![\\w.,–-])/(?:bin|boot|dev|etc|home|lib|mnt|opt|root|sbin|tmp|usr|var|Bureau|Documents|Images|Musique|Public|Téléchargements|Vidéos)(?:/[\\w.()]+)*', False, '#600p', 'p_arborescence_Linux_Mac', 4, [[None, '~', '*', 0]]],
    ['(?i)(?<![\\w.,–-])[a-z]:\\\\(?:Program Files(?: [(]x86[)]|)|[\\w.()]+)(?:\\\\[\\w.()]+)*', True, '#603p', 'p_arborescence_Windows', 4, [[None, '~', '*', 0]]],
    ['(?<![\\w.,–-])\\d+ (m) ', False, '#607p', 'p_chiffres_m', 4, [[None, '~', '_', 1]]],
  ]],
  ["apos", [
    ['(?i)(?<![\\w.,–-])([ldsncjmç]|jusqu|lorsqu|aujourd|presqu|quelqu|puisqu|qu|prud|entr)[\'ʼ‘‛´`′‵՚ꞌꞋ](?=[\\w"«<])', True, '#644p', 'apostrophe_typographique', 8, [[None, '-', '\\1’', 0, 'Apostrophe typographique.', '']]],
    ['(?i)(?<![\\w.,–-])t[\'ʼ‘‛´`′‵՚ꞌꞋ](?!ils?|elles?|on)(?=[\\w"«<])', True, '#646p', 'apostrophe_typographique_après_t', 8, [[None, '-', 't’', 0, 'Apostrophe typographique.', '']]],
  ]],
  ["typo", [
    ['(?i)(?<![\\w.,–-])(([ldsncjmtç]|lorsqu|puisqu|presqu|quoiqu|quelqu|jusqu|qu|aujourd|entr)(?:[\'’ʼ‘‛´`′‵՚ꞌꞋ][\'’ʼ‘‛´`′‵՚ꞌꞋ ]|”|“|"| [\'’ʼ‘‛´`′‵՚ꞌꞋ] ?))[aeéiouhœæyàîèêôû<]', True, '#655p', 'typo_apostrophe_incorrecte', 8, [['_c_typo_apostrophe_incorrecte_1', '-', '\\2’', 1, 'Apostrophe incorrecte.', '']]],
    ['(?i)(?<![\\w.,–-])((lorsqu|puisqu|presqu|quoiqu|jusqu|qu|aujourd|entr) )[aeéiouhyàîèêôû<]', True, '#666p', 'typo_apostrophe_manquante', 8, [[None, '-', '\\2’', 1, 'Il manque vraisemblablement une apostrophe.', ''], [None, '~', '\\2’', 1]]],
    ['(?i)(?<![\\w.,–-])ne ([mtsl] )[aeéiouhyàîèêôû<]', True, '#671p', 'typo_apostrophe_manquante_prudence1', 8, [[None, '-', '=_s_typo_apostrophe_manquante_prudence1_1', 1, 'Il manque une apostrophe.', '']]],
    ['(?i)(?<![\\w.,–-])(?:je|tu|ie?ls?|nous|vous|on|ça|elles?) ([nmtsl] )([aeéiouhyàîèêôû][\\w-]*|<)', True, '#674p', 'typo_apostrophe_manquante_prudence2', 8, [['_c_typo_apostrophe_manquante_prudence2_1', '-', '=_s_typo_apostrophe_manquante_prudence2_1', 1, 'Il manque probablement une apostrophe.', '']]],
    ['(?<![\\w.,–-])([ldsncjmtç] )[aeéiouhAEÉIOUHyîèêôûYÎÈÊÔÛ<]', False, '#678p', 'typo_apostrophe_manquante_audace1', 8, [['_c_typo_apostrophe_manquante_audace1_1', '-', '=_s_typo_apostrophe_manquante_audace1_1', 1, 'Il manque peut-être une apostrophe.', '']]],
    ['"(?=\\w)', False, '#702p', 'typo_guillemets_typographiques_doubles_ouvrants', 4, [['_c_typo_guillemets_typographiques_doubles_ouvrants_1', '-', '«\xa0|“', 0, 'Guillemets typographiques ouvrants.', '']]],
    ['[\\w.?!…,](")(?!\\w)', False, '#705p', 'typo_guillemets_typographiques_doubles_fermants', 4, [[None, '-', '\xa0»|”', 1, 'Guillemets typographiques fermants.', '']]],
    ["(?:^|[ –—-])(['’‘]['’‘])(?=\\w)", False, '#713p', 'typo_guillemets_typographiques_simples_doubles_ouvrants', 4, [[None, '-', '«\xa0|“', 1, 'Guillemets typographiques ouvrants.', '']]],
    ["[\\w.?!…,](['’]['’])(?![\\w'’])", False, '#716p', 'typo_guillemets_typographiques_simples_doubles_fermants', 4, [[None, '-', '\xa0»|”', 1, 'Guillemets typographiques fermants.', '']]],
    ["(?:^|[ –—-])(['’])(?=\\w)", False, '#726p', 'typo_guillemets_typographiques_simples_ouvrants', 4, [[None, '-', '“|‘|‹', 1, 'Guillemets typographiques ouvrants.', '']]],
    ["[\\w.?!](')(?:[ \xa0\u202f]|$)", False, '#729p', 'typo_guillemets_typographiques_simples_fermants', 4, [[None, '-', '”|’|›', 1, 'Guillemets typographiques fermants.', '']]],
  ]],
  ["html", [
    ['(?i)<\\w+.*?>', True, '#748p', 'p_html_balise_ouvrante', 4, [[None, '~', '*', 0]]],
    ['(?i)</\\w+ *>', True, '#749p', 'p_html_balise_fermante', 4, [[None, '~', '*', 0]]],
    ['(?i)&amp;[a-zA-Z]+;', True, '#750p', 'p_html_amp_xxx', 4, [[None, '~', '_', 0]]],
    ['(?i)&lt;', True, '#751p', 'p_html_lt', 4, [[None, '~', '   <', 0]]],
    ['(?i)&gt;', True, '#752p', 'p_html_gt', 4, [[None, '~', '>', 0]]],
    ['(?i)&amp;', True, '#753p', 'p_html_amp', 4, [[None, '~', '&', 0]]],
    ['(?i)&nbsp;', True, '#754p', 'p_html_nbsp', 4, [[None, '~', '@', 0]]],
    ['(?i)&#(?:160|8239);', True, '#755p', 'p_html_nbsp2', 4, [[None, '~', '@', 0]]],
    ['(?i)\\[(?:[buiqa]|blockquote|quote)\\]', True, '#756p', 'p_html_pseudo_balise1', 4, [[None, '~', '*', 0]]],
    ['(?i)\\[/\\w+\\]', True, '#757p', 'p_html_pseudo_balise2', 4, [[None, '~', '*', 0]]],
  ]],
  ["latex", [
    ['(?i)\\\\[a-z]+(?![\\w–-])', True, '#765p', 'p_latex1', 4, [[None, '~', '*', 0]]],
    ['(?i)\\\\[,;/\\\\]', True, '#766p', 'p_latex2', 4, [[None, '~', '*', 0]]],
    ['\\{(?:abstract|align|cases|center|description|enumerate|equation|figure|flush(?:left|right)|gather|minipage|multline|quot(?:ation|e)|SaveVerbatim|table|tabular|thebibliography|[vV]erbatim|verse|wrapfigure)\\}', False, '#767p', 'p_latex3', 4, [[None, '~', '*', 0]]],
  ]],
  ["md", [
    ['(?:[ \u202f\xa0]|^)([*_]+)\\w', False, '#770p', 'p_md_span_ouvrant', 4, [[None, '~', '*', 1]]],
    ['(?i)[a-z0-9àéîïôûüù]([*_]+)(?:[,.;:!?) \u202f\xa0]|$)', True, '#771p', 'p_md_span_fermant', 4, [[None, '~', '*', 1]]],
    ['^#+', False, '#772p', 'p_md_titre_début', 4, [[None, '~', '*', 0]]],
    ['#+ *$', False, '#773p', 'p_md_titre_fin', 4, [[None, '~', '*', 0]]],
    ['^ *[*+] ', False, '#774p', 'p_md_liste', 4, [[None, '~', '*', 0]]],
    ['^>[ >]+ ', False, '#775p', 'p_md_citation', 4, [[None, '~', '*', 0]]],
  ]],
  ["", [
    ['(?i)(?<![\\w.,–-])tous?[.(/·•⋅–—-]te[.)/·•⋅–—-]?s(?![\\w–-])', True, '#793p', 'eepi_écriture_épicène_tous_toutes', 4, [['_c_eepi_écriture_épicène_tous_toutes_1', '-', 'tous et toutes|toutes et tous', 0, 'Écriture épicène dystypographique et imprononçable. À des fins de lisibilité, il est préférable d’éviter l’abus de graphies épicènes trop complexes.', ''], [None, '~', '=_p_eepi_écriture_épicène_tous_toutes_2', 0]]],
    ['(?i)(?<![\\w.,–-])c[./·•⋅–—-]?eux?[./·•⋅–—-]elles(?![\\w–-])', True, '#803p', 'eepi_écriture_épicène_ceux_celles', 4, [['_c_eepi_écriture_épicène_ceux_celles_1', '-', 'ceux et celles|celles et ceux', 0, 'Écriture épicène dystypographique et imprononçable. À des fins de lisibilité, il est préférable d’éviter l’abus de graphies épicènes trop complexes.', ''], [None, '~', '=_p_eepi_écriture_épicène_ceux_celles_2', 0]]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]+)eurs?[.(/·•⋅–—-][tdp]?(rice|euse|se|[oe]resse)[.)/·•⋅–—-]?s(?![\\w–-])', True, '#813p', 'eepi_écriture_épicène_pluriel_eur_divers', 4, [['_c_eepi_écriture_épicène_pluriel_eur_divers_1', '-', '\\1eurs et \\1\\2s|\\1\\2s et \\1eurs', 0, 'Écriture épicène dystypographique et imprononçable. À des fins de lisibilité, il est préférable d’éviter l’abus de graphies épicènes trop complexes.', ''], ['_c_eepi_écriture_épicène_pluriel_eur_divers_2', '-', '\\1eurs et \\1euses|\\1euses et \\1eurs', 0, 'Écriture épicène dystypographique et imprononçable. À des fins de lisibilité, il est préférable d’éviter l’abus de graphies épicènes trop complexes.', ''], [None, '~', '=_p_eepi_écriture_épicène_pluriel_eur_divers_3', 0]]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]+)eux[.(/·•⋅–—-](?:[tdsi]?euse|se)[.)/·•⋅–—-]?s(?![\\w–-])', True, '#825p', 'eepi_écriture_épicène_pluriel_eux_euses', 4, [['_c_eepi_écriture_épicène_pluriel_eux_euses_1', '-', '\\1eux et \\1euses|\\1euses et \\1eux', 0, 'Écriture épicène dystypographique et imprononçable. À des fins de lisibilité, il est préférable d’éviter l’abus de graphies épicènes trop complexes.', ''], [None, '~', '=_p_eepi_écriture_épicène_pluriel_eux_euses_2', 0]]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]+)ifs?[.(/·•⋅–—-][std]?i?ve[.)/·•⋅–—-]?s(?![\\w–-])', True, '#834p', 'eepi_écriture_épicène_pluriel_if_ive', 4, [['_c_eepi_écriture_épicène_pluriel_if_ive_1', '-', '\\1ifs et \\1ives|\\1ives et \\1ifs', 0, 'Écriture épicène dystypographique et imprononçable. À des fins de lisibilité, il est préférable d’éviter l’abus de graphies épicènes trop complexes.', ''], [None, '~', '=_p_eepi_écriture_épicène_pluriel_if_ive_2', 0]]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]+)[eè]rs?[.(/·•⋅–—-](?:i?è?re|e)[.)/·•⋅–—-]?s(?![\\w–-])', True, '#844p', 'eepi_écriture_épicène_pluriel_er_ère', 4, [['_c_eepi_écriture_épicène_pluriel_er_ère_1', '-', '\\1er·e·s|\\1ers et \\1ères|\\1ères et \\1ers', 0, 'Écriture épicène dystypographique et imprononçable. À des fins de lisibilité, il est préférable d’éviter l’abus de graphies épicènes trop complexes.', ''], [None, '~', '=_p_eepi_écriture_épicène_pluriel_er_ère_2', 0]]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]+)aux[.(/·•⋅–—-][tnmcpbd]?a?le[.)/·•⋅–—-]?s(?![\\w–-])', True, '#856p', 'eepi_écriture_épicène_pluriel_aux_ales', 4, [['_c_eepi_écriture_épicène_pluriel_aux_ales_1', '-', '\\1al·e·s|\\1aux et \\1ales|\\1ales et \\1aux', 0, 'Écriture épicène dystypographique et imprononçable. Pour ce cas, il peut être intéressant de faire comme si le pluriel masculin était régulier, ce qui rend l’ensemble prononçable…', ''], [None, '~', '=_p_eepi_écriture_épicène_pluriel_aux_ales_2', 0]]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]*[éuitsnldf])[-·–—.•⋅(/]([ntlf]?e)[-·–—.•⋅)/]?s(?![\\w–-])', True, '#867p', 'eepi_écriture_épicène_pluriel_e', 4, [['_c_eepi_écriture_épicène_pluriel_e_1', '>', ''], [None, '~', '=_p_eepi_écriture_épicène_pluriel_e_2', 0], ['_c_eepi_écriture_épicène_pluriel_e_3', '>', ''], ['_c_eepi_écriture_épicène_pluriel_e_4', '-', '\\1·\\2·s|\\1 et \\1\\2s|\\1\\2s et \\1', 0, 'Écriture épicène. Utilisez les points médians ou écrivez en toutes lettres.', ''], ['_c_eepi_écriture_épicène_pluriel_e_5', '-', '\\1·\\2·s|\\1s et \\1\\2s|\\1\\2s et \\1s', 0, 'Écriture épicène. Utilisez les points médians ou écrivez en toutes lettres.', '']]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]*[éuitsrnldf])([-·–—.•⋅/][ntl]?e|[(][ntl]?e[)])(?![\\w–-])', True, '#897p', 'eepi_écriture_épicène_singulier', 4, [['_c_eepi_écriture_épicène_singulier_1', '>', ''], [None, '~', '=_p_eepi_écriture_épicène_singulier_2', 0], ['_c_eepi_écriture_épicène_singulier_3', '-', 'un·e|un ou une|une ou un', 0, 'Écriture épicène. Utilisez les points médians ou écrivez en toutes lettres.', ''], ['_c_eepi_écriture_épicène_singulier_4', '-', '=_s_eepi_écriture_épicène_singulier_4', 0, 'Écriture épicène. Utilisez un point médian.', '']]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]+\\w)([-·–—.•⋅/]s|[(]s[)])(?![\\w–-])', True, '#917p', 'typo_écriture_invariable', 4, [['_c_typo_écriture_invariable_1', '~', '=_p_typo_écriture_invariable_1', 0], ['_c_typo_écriture_invariable_2', '-', '\\1·s', 0, 'Écriture invariable. Utilisez un point médian.', '']]],
  ]],
  ["maj", [
    ['(?<![\\w.,–-])(\\w[\\w-]+)[.] ([a-zàâéèêîôç]\\w*)(?![\\w–-])', False, '#934p', 'majuscule_après_point', 4, [['_c_majuscule_après_point_1', '-', '=_s_majuscule_après_point_1', 2, 'Après un point, une majuscule est généralement requise.', '']]],
    ["^ *([a-zàâéèêîôç](?:[\\w-]+[’'`‘]?|[’'`‘]))", False, '#943p', 'majuscule_début_paragraphe', 4, [['_c_majuscule_début_paragraphe_1', '-', '=_s_majuscule_début_paragraphe_1', 1, 'Majuscule en début de phrase, sauf éventuellement lors d’une énumération.', '']]],
  ]],
  ["poncfin", [
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]*) *$', True, '#959p', 'poncfin_règle1', 4, [['_c_poncfin_règle1_1', '-', '\\1.|\\1\xa0!|\\1\xa0?', 1, 'Il semble manquer une ponctuation finale (s’il s’agit d’un titre, le point final n’est pas requis).', '']]],
  ]],
  ["typo", [
    ['(\\w+)[,:](\\w[\\w-]*)', False, '#977p', 'typo_espace_manquant_après1', 4, [['_c_typo_espace_manquant_après1_1', '-', ' \\2', 2, 'Il manque un espace.', '']]],
    ['[?!;%‰‱˚»}](\\w[\\w-]*|[({[])', False, '#979p', 'typo_espace_manquant_après2', 4, [[None, '-', ' \\1', 1, 'Il manque un espace.', '']]],
    ['(?<![\\w.,–-])\\w[\\w-]*[.](\\w[\\w-]*|[({[])', False, '#981p', 'typo_espace_manquant_après3', 4, [['_c_typo_espace_manquant_après3_1', '-', ' \\1', 1, 'Il manque un espace.', '']]],
    ['(?<![\\w.,–-])\\w[\\w-]+[…](\\w[\\w-]*|[({[])', False, '#984p', 'typo_espace_manquant_après4', 4, [['_c_typo_espace_manquant_après4_1', '-', ' \\1', 1, 'Il manque un espace.', '']]],
    ['(?i)(?<![\\w.,–-])etc(?:[.]{3,5}|…)', True, '#1018p', 'typo_et_cetera', 4, [[None, '-', 'etc.', 0, 'Un seul point après «\xa0etc.\xa0»', '']]],
    ['(?<![\\w.,–-])(M(?:me|gr|lle)s?[.]) [A-ZÉÈ]', False, '#1024p', 'typo_point_après_titre', 4, [[None, '-', '=_s_typo_point_après_titre_1', 1, 'Pas de point après cette abréviation.', '']]],
    ['(?<![\\w.,–-])([nN]os?[.])[ \xa0]\\d+', False, '#1030p', 'typo_point_après_numéro', 4, [['_c_typo_point_après_numéro_1', '-', 'nᵒˢ', 1, 'Pas de point dans l’abréviation de numéro.', ''], ['_c_typo_point_après_numéro_2', '-', 'nᵒ', 1, 'Pas de point dans l’abréviation de numéro.', '']]],
    ['(?<![.])[.]{3,5}(?![.])', False, '#1039p', 'typo_points_suspension1', 4, [['_c_typo_points_suspension1_1', '-', '…', 0, 'Typographie\xa0: points de suspension. Utilisez le caractère dédié.', '']]],
    ['[.]{6,}', False, '#1040p', 'typo_points_suspension2', 4, [[None, '-', '=_s_typo_points_suspension2_1', 0, 'Typographie\xa0: succession de points. Utilisez le caractère dédié.', '']]],
    ['(?<![.…])[.][.](?![.])', False, '#1041p', 'typo_points_suspension3', 4, [[None, '-', '…|.', 0, 'Typographie\xa0: un ou trois points\xa0?', '']]],
    ['…[.][.]?(?![.])', False, '#1042p', 'typo_points_superflus', 4, [[None, '-', '…', 0, 'Points superflus.', '']]],
    [',[.,]{2,}', False, '#1062p', 'typo_virgules_points', 4, [[None, '-', '=_s_typo_virgules_points_1', 0, 'Erreur de numérisation\xa0? Virgules au lieu de points\xa0?', '']]],
    [', ?([.,;:!?])', False, '#1069p', 'typo_ponctuation_superflue1', 4, [[None, '-', '=_s_typo_ponctuation_superflue1_1', 0, 'Une de ces ponctuations est superflue.', '']]],
    ['; ?([.,;:…!?])', False, '#1070p', 'typo_ponctuation_superflue2', 4, [[None, '-', '=_s_typo_ponctuation_superflue2_1', 0, 'Une de ces ponctuations est superflue.', '']]],
    [':[.,;:…!?]', False, '#1071p', 'typo_ponctuation_superflue3', 4, [[None, '-', '=_s_typo_ponctuation_superflue3_1', 0, 'Une de ces ponctuations est superflue.', '']]],
    ['[?!][.](?![.])', False, '#1072p', 'typo_ponctuation_superflue4', 4, [[None, '-', '=_s_typo_ponctuation_superflue4_1', 0, 'Ne mettez pas un point après un point d’interrogation ou d’exclamation.', '']]],
  ]],
  ["nbsp", [
    ['\\b[?!;]', False, '#1087p', 'nbsp_ajout_avant_double_ponctuation', 4, [['_c_nbsp_ajout_avant_double_ponctuation_1', '-', '=_s_nbsp_ajout_avant_double_ponctuation_1', 0, 'Il manque un espace insécable.', '']]],
    [' ([?!;])', False, '#1089p', 'nbsp_avant_double_ponctuation', 4, [[None, '-', '\xa0\\1', 0, 'Il manque un espace insécable.', '']]],
    ['(?: |\\b):(?= |$)', False, '#1091p', 'nbsp_avant_deux_points', 4, [[None, '-', '\xa0:', 0, 'Il manque un espace insécable.', '']]],
    ['«(?: |\\b)', False, '#1093p', 'nbsp_après_chevrons_ouvrants', 4, [[None, '-', '«\xa0', 0, 'Il manque un espace insécable.', '']]],
    ['^»(?: |\\b)', False, '#1095p', 'nbsp_après_chevrons_fermants', 4, [[None, '-', '»\xa0|«\xa0', 0, 'Il manque un espace insécable. Le sens de ce guillemet n’est justifié que si ce paragraphe continue le discours du paragraphe précédent.', '']]],
    ['(?: |\\b)»', False, '#1097p', 'nbsp_avant_chevrons_fermants1', 4, [[None, '-', '\xa0»', 0, 'Il manque un espace insécable.', '']]],
    ['([].!?)])»', False, '#1099p', 'nbsp_avant_chevrons_fermants2', 4, [[None, '-', '\\1\xa0»', 0, 'Il manque un espace insécable.', '']]],
    ['([:;!?])\xa0(?=[\\w«“"])', False, '#1101p', 'nbsp_après_double_ponctuation', 4, [[None, '-', '\\1 ', 0, 'Pas d’espace insécable après le signe “\\1”.', '']]],
  ]],
  ["unit", [
    ["(?<![\\w.,–-])((\\d+(?:,\\d+[⁰¹²³⁴⁵⁶⁷⁸⁹]?|[⁰¹²³⁴⁵⁶⁷⁸⁹]|)) ?)([kcmµn]?(?:[slgJKVΩ]|m[²³]?|Wh?|Hz|dB)|[%‰€$£¥Åℓhj]|min|jours?|semaines|mois|ans?|euros?|heures?|px|MHz|°C|℃)(?![’'])(?![\\w–-])", False, '#1117p', 'unit_nbsp_avant_unités1', 4, [['_c_unit_nbsp_avant_unités1_1', '-', '=_s_unit_nbsp_avant_unités1_1', 0, 'Avec une unité de mesure, mettez un espace insécable.', ''], ['_c_unit_nbsp_avant_unités1_2', '-', '\\2\xa0\\3', 0, 'Avec une unité de mesure, mettez un espace insécable.', '']]],
    ['(?<![\\w.,–-])((\\d+(?:,\\d+[⁰¹²³⁴⁵⁶⁷⁸⁹]?|[⁰¹²³⁴⁵⁶⁷⁸⁹])) ?)([a-zA-Zµ][a-zA-Z0-9Ωℓ⁰¹²³⁴⁵⁶⁷⁸⁹/·]*)(?![\\w–-])', False, '#1122p', 'unit_nbsp_avant_unités2', 4, [['_c_unit_nbsp_avant_unités2_1', '>', ''], ['_c_unit_nbsp_avant_unités2_2', '-', '=_s_unit_nbsp_avant_unités2_2', 0, 'Si ‹\\3› est une unité de mesure, il manque un espace insécable. Si le nombre se rapporte au mot suivant, c’est aussi valable.', ''], ['_c_unit_nbsp_avant_unités2_3', '-', '\\2\xa0\\3', 0, 'Si ‹\\3› est une unité de mesure, il manque un espace insécable. Si le nombre se rapporte au mot suivant, c’est aussi valable.', '']]],
    ["(?<![\\w.,–-])((\\d+) )([a-zA-Zµ][a-zA-Z0-9Ωℓ⁰¹²³⁴⁵⁶⁷⁸⁹/·]*)(?![’'])(?![\\w–-])", False, '#1128p', 'unit_nbsp_avant_unités3', 4, [['_c_unit_nbsp_avant_unités3_1', '>', ''], ['_c_unit_nbsp_avant_unités3_2', '-', '=_s_unit_nbsp_avant_unités3_2', 0, 'Si ‹\\3› est une unité de mesure, il manque un espace insécable. Si le nombre se rapporte au mot suivant, c’est aussi valable.', ''], ['_c_unit_nbsp_avant_unités3_3', '-', '\\2\xa0\\3', 0, 'Si ‹\\3› est une unité de mesure, il manque un espace insécable. Si le nombre se rapporte au mot suivant, c’est aussi valable.', '']]],
  ]],
  ["typo", [
    ["(?<![\\w.,–-])(?:cos|sin|tan|cotan|log|ln)²? [2-9]?[a-zA-Zα-ωΑ-Ω](?![’'])(?![\\w–-])", False, '#1159p', 'typo_math', 4, [[None, '-', '=_s_typo_math_1', 0, 'Mettez des parenthèses (ou éventuellement un espace insécable).', '']]],
    ['^[-_][ \xa0]', False, '#1175p', 'typo_tiret_début_ligne', 4, [[None, '-', '—\xa0|–\xa0', 0, 'Dialogues et énumérations\xa0: un tiret cadratin [1] ou demi-cadratin [2], suivi d’un espace insécable, est requis.', '']]],
    ['^ *«[ \xa0](-[ \xa0])', False, '#1176p', 'typo_tiret_dans_dialogue', 4, [[None, '-', '—\xa0|–\xa0', 1, 'Dialogues\xa0: un tiret cadratin [1] ou demi-cadratin [2], suivi d’un espace insécable, est requis. (Attention\xa0: à moins qu’il s’agisse d’un dialogue inclus dans un autre dialogue, cette manière d’écrire est erronée.)', '']]],
  ]],
  ["nbsp", [
    ['^([—–]) +', False, '#1178p', 'nbsp_après_tiret1', 4, [[None, '-', '\\1\xa0', 0, 'Dialogues et énumérations\xa0: ce tiret doit être suivi d’un espace insécable.', '']]],
    ['^([—–-])\\d', False, '#1179p', 'nbsp_après_tiret2', 4, [[None, '-', '—\xa0|–\xa0|−', 1, 'Dialogues et énumérations\xa0: il faut tiret cadratin [1] ou demi-cadratin [2] suivi d’un espace insécable. S’il s’agit de représenter un nombre négatif, utilisez le signe “moins” [3].', '']]],
    ['^([—–-])[\\w«“‘]', False, '#1180p', 'nbsp_après_tiret3', 4, [[None, '-', '—\xa0|–\xa0', 1, 'Dialogues et énumérations\xa0: il faut tiret cadratin [1] ou demi-cadratin [2] suivi d’un espace insécable.', '']]],
  ]],
  ["typo", [
    [' - (?=[0-9])', False, '#1181p', 'typo_tiret_incise1', 4, [[None, '-', ' – | — | − | −', 0, 'Tiret d’incise\xa0: un tiret demi-cadratin [1] ou cadratin [2] est requis. (Ou éventuellement le signe “moins” s’il s’agit d’une soustraction [3] ou de représenter un nombre négatif [4].)', '']]],
    [' - (?![0-9])', False, '#1182p', 'typo_tiret_incise2', 4, [[None, '-', ' – | — | − ', 0, 'Tiret d’incise\xa0: un tiret demi-cadratin [1] ou cadratin [2] est requis. (Ou éventuellement le signe “moins” [3], s’il s’agit d’une soustraction.)', '']]],
    ['-(\\d+)(?![\\w–-])', False, '#1201p', 'typo_signe_moins', 4, [['_c_typo_signe_moins_1', '-', '−\\1|– \\1|— \\1', 0, 'S’il s’agit de représenter un nombre négatif, utilisez le signe typographique “moins” [1]. S’il s’agit d’une incise, utilisez un tiret demi-cadratin [2] ou un tiret cadratin [3].', '']]],
    ['(\\d+) ?[x*] ?(\\d+)', False, '#1208p', 'typo_signe_multiplication', 4, [['_c_typo_signe_multiplication_1', '-', '\\1\xa0×\xa0\\2', 0, 'Signe de multiplication typographique.', '']]],
    ['[\\w \xa0](<=)[\\w \xa0]', False, '#1217p', 'typo_inférieur_ou_égal', 4, [[None, '-', '≤|⩽', 1, 'Signe “inférieur ou égal”.', '']]],
    ['[\\w \xa0](>=)[\\w \xa0]', False, '#1218p', 'typo_supérieur_ou_égal', 4, [[None, '-', '≥|⩾', 1, 'Signe “supérieur ou égal”.', '']]],
  ]],
  ["liga", [
    ['[ﬁﬂﬀﬃﬄﬅﬆ]', False, '#1225p', 'ligatures_typographiques', 4, [[None, '-', '=_s_ligatures_typographiques_1', 0, 'Suppression de la ligature typographique.', '']]],
  ]],
  ["nf", [
    ['(?<![\\w.,–-])[nN][fF][ \xa0-]?(?:c|C|e|E|p|P|q|Q|s|S|x|X|z|Z|[eE][nN](?:[ \xa0-][iI][sS][oO]|))[ \xa0-]?[0-9]+(?:[/ ‑-][0-9]+|)(?![\\w–-])', False, '#1241p', 'nf_norme_française', 4, [['_c_nf_norme_française_1', '-', '=_s_nf_norme_française_1', 0, 'Norme française. Utilisez les espaces et tirets insécables.', 'http://fr.wikipedia.org/wiki/Liste_de_normes_NF']]],
  ]],
  ["typo", [
    ['(«[\xa0\u202f])[^»“]+?(”)', False, '#1256p', 'typo_cohérence_guillemets_chevrons_ouvrants', 4, [['_c_typo_cohérence_guillemets_chevrons_ouvrants_1', '-', '“', 1, 'Guillemet ouvrant différent du guillemet fermant (”).', ''], ['_c_typo_cohérence_guillemets_chevrons_ouvrants_2', '-', '\xa0»', 2, 'Guillemet fermant différent du guillemet ouvrant («).', '']]],
    ['(“)[^”«]+?([\xa0\u202f]»)', False, '#1260p', 'typo_cohérence_guillemets_chevrons_fermants', 4, [['_c_typo_cohérence_guillemets_chevrons_fermants_1', '-', '«\xa0', 1, 'Guillemet ouvrant différent du guillemet fermant (»).', ''], ['_c_typo_cohérence_guillemets_chevrons_fermants_2', '-', '”', 2, 'Guillemet fermant différent du guillemet ouvrant (“).', '']]],
    ["(“)[^”’“«]+?(’)(?![\\w'’)])", False, '#1269p', 'typo_cohérence_guillemets_doubles_ouvrants', 4, [['_c_typo_cohérence_guillemets_doubles_ouvrants_1', '>', ''], ['_c_typo_cohérence_guillemets_doubles_ouvrants_2', '-', '‘', 1, 'Guillemet ouvrant différent du guillemet fermant (’).', ''], ['_c_typo_cohérence_guillemets_doubles_ouvrants_3', '-', '”', 2, 'Guillemet fermant différent du guillemet ouvrant (“).', '']]],
    ['(‘)[^’“«]+?(”)', False, '#1274p', 'typo_cohérence_guillemets_doubles_fermants', 4, [['_c_typo_cohérence_guillemets_doubles_fermants_1', '-', '“', 1, 'Guillemet ouvrant différent du guillemet fermant (”).', ''], ['_c_typo_cohérence_guillemets_doubles_fermants_2', '-', '’', 2, 'Guillemet fermant différent du guillemet ouvrant (‘).', '']]],
    ['(?i)(?<![\\w.,–-])(“)\\w[^”"»]+$', True, '#1285p', 'typo_guillemets_doubles_ouvrants_non_fermés', 4, [[None, '-', '_', 1, 'Guillemets fermants introuvables dans la suite du paragraphe.', '']]],
    ['(?i)^[^“„«"]+\\w(”)(?![\\w–-])', True, '#1288p', 'typo_guillemets_doubles_fermants_non_ouverts', 4, [[None, '-', '_', 1, 'Guillemets ouvrants introuvables dans ce paragraphe.', '']]],
    ["(?i)(?<![\\w.,–-])(‘)\\w\\w[^’']+$", True, '#1296p', 'typo_guillemet_simple_ouvrant_non_fermé', 4, [['_c_typo_guillemet_simple_ouvrant_non_fermé_1', '-', '_', 1, 'Guillemet fermant introuvable dans la suite du paragraphe.', '']]],
    ["(?i)^[^‘']+\\w\\w(’)(?![\\w–-])", True, '#1299p', 'typo_guillemet_simple_fermant_non_ouvert', 4, [['_c_typo_guillemet_simple_fermant_non_ouvert_1', '-', '_', 1, 'Guillemet ouvrant introuvable dans ce paragraphe.', '']]],
  ]],
  ["num", [
    ['(?<![\\w.,–-])\\d\\d\\d\\d+(?![\\w–-])', False, '#1317p', 'num_grand_nombre_soudé', 4, [['_c_num_grand_nombre_soudé_1', '>', ''], ['_c_num_grand_nombre_soudé_2', '-', '=_s_num_grand_nombre_soudé_2', 0, 'Formatage des grands nombres.', ''], ['_c_num_grand_nombre_soudé_3', '-', '=_s_num_grand_nombre_soudé_3', 0, 'Formatage des grands nombres.', '']]],
    ["(?<![\\w.,–-])(?<!,)(\\d\\d\\d\\d)[\xa0\u202f]([a-zA-Zµ][a-zA-Z0-9Ωℓ⁰¹²³⁴⁵⁶⁷⁸⁹/·]*)(?![’'])(?![\\w–-])", False, '#1325p', 'num_nombre_quatre_chiffres', 4, [['_c_num_nombre_quatre_chiffres_1', '-', '=_s_num_nombre_quatre_chiffres_1', 1, 'Formatage des grands nombres.', '']]],
  ]],
  ["", [
    ['(?<![\\w.,–-])\\d\\d?\\d?(?:[.]\\d\\d\\d)+(?![0-9])', False, '#1352p', 'num_grand_nombre_avec_points', 4, [['_c_num_grand_nombre_avec_points_1', '-', '=_s_num_grand_nombre_avec_points_1', 0, 'Grands nombres\xa0: utilisez des espaces insécables plutôt que des points.', ''], [None, '~', '=_p_num_grand_nombre_avec_points_2', 0]]],
    ['(?<![\\w.,–-])\\d\\d?\\d?(?: \\d\\d\\d)+(?![\\w–-])', False, '#1360p', 'num_grand_nombre_avec_espaces', 4, [['_c_num_grand_nombre_avec_espaces_1', '-', '=_s_num_grand_nombre_avec_espaces_1', 0, 'Grands nombres\xa0: utilisez des espaces insécables.', ''], [None, '~', '=_p_num_grand_nombre_avec_espaces_2', 0]]],
  ]],
  ["date", [
    ['(?i)(?<![\\w.,–-])(?<!\\d[ ./-])(\\d\\d?)([ ./-])(\\d\\d?)([ ./-])(\\d\\d\\d+)(?![ ./-]\\d)(?![\\w–-])', True, '#1373p', 'date_nombres', 4, [['_c_date_nombres_1', '-', '_', 0, 'Cette date est invalide.', ''], [None, '~', '\\1-\\3-\\5', 0]]],
  ]],
  ["redon1", [
    ['(?i)(?<![\\w.,–-])(\\w[\\w-][\\w-]+\\w)[ \xa0,.;!?:].*[ \xa0](\\1)(?![\\w–-])', True, '#1392p', 'redondances_paragraphe', 4, [['_c_redondances_paragraphe_1', '-', '_', 2, 'Dans ce paragraphe, répétition de «\xa0\\1\xa0» (à gauche).', ''], ['_c_redondances_paragraphe_2', '-', '_', 1, 'Dans ce paragraphe, répétition de «\xa0\\1\xa0» (à droite).', '']]],
  ]],
  ["ocr", [
    ['[ \xa0]7(?![\\w–-])', False, '#1410p', 'ocr_point_interrogation', 3, [['_c_ocr_point_interrogation_1', '-', '\xa0?', 0, 'Erreur de numérisation\xa0?', '']]],
    ["[ \xa0]I(?![ ’'][aâeéèêëiîïoôuy])(?![\\w–-])", False, '#1419p', 'ocr_exclamation1', 3, [[None, '-', '\xa0!', 0, 'Erreur de numérisation\xa0?', '']]],
    ['[ \xa0]1(?= [A-ZÉÈÂÎ])(?![\\w–-])', False, '#1422p', 'ocr_exclamation2', 3, [['_c_ocr_exclamation2_1', '-', '\xa0!', 0, 'Erreur de numérisation\xa0?', '']]],
    ['(?i)(?<![\\w.,–-])[\\dOI][\\dOI]+(?![\\w–-])', True, '#1433p', 'ocr_nombres', 3, [['_c_ocr_nombres_1', '-', '11|Il', 0, 'Erreur de numérisation\xa0?', ''], ['_c_ocr_nombres_2', '-', '=_s_ocr_nombres_2', 0, 'Erreur de numérisation\xa0?', '']]],
    ['(?<![\\w.,–-])\\w\\w+-(On|Ils?|Elles?|Tu|Je|Nous|Vous|Mêmes?|Ci|Là|Une?s|Les?|La|Leur)(?![\\w–-])', False, '#1443p', 'ocr_casse_pronom_vconj', 3, [[None, '-', '=_s_ocr_casse_pronom_vconj_1', 1, 'Erreur de numérisation\xa0? Casse douteuse.', '']]],
    ['(?<![\\w.,–-])\\w+-\\w+(?![\\w–-])', False, '#1453p', 'mots_composés_inconnus', 3, [['_c_mots_composés_inconnus_1', '-', '_', 0, 'Erreur de numérisation\xa0? Mot composé inconnu du dictionnaire.', '']]],
    ['(?i)(?<![\\w.,–-])(\\w+)‑(\\w+)(?![\\w–-])', True, '#1467p', 'tu_trait_union_conditionnel', 3, [[None, '-', '\\1\\2|\\1-\\2', 0, 'Trait d’union conditionnel. Erreur de numérisation\xa0?', '']]],
    ['\\w*[{}<>&*#£^|]+\\w*', False, '#1475p', 'ocr_caractères_rares', 3, [['_c_ocr_caractères_rares_1', '-', '_', 0, 'Erreur de numérisation\xa0? Cette chaîne contient un caractère de fréquence rare.', '']]],
    ['[\\]\\[({}][\\]\\[({}]+', False, '#1485p', 'ocr_doublons_caractères_rares', 3, [[None, '-', '_', 0, 'Erreur de numérisation\xa0? Succession\xa0douteuse de caractères.', '']]],
    ['(?i)(?<![\\w.,–-])[1[\\]][easrnxiocuwàéè](?![\\w–-])', True, '#1494p', 'ocr_le_la_les_regex', 3, [['_c_ocr_le_la_les_regex_1', '-', 'le', 0, 'Erreur de numérisation\xa0?', ''], ['_c_ocr_le_la_les_regex_2', '-', 'la', 0, 'Erreur de numérisation\xa0?', ''], ['_c_ocr_le_la_les_regex_3', '-', 'la|là', 0, 'Erreur de numérisation\xa0?', ''], ['_c_ocr_le_la_les_regex_4', '-', 'le|la', 0, 'Erreur de numérisation\xa0?', '']]],
  ]],
  ["conf", [
    ['(?i)(?<![\\w.,–-])[1[\\]][ea]s?(?![\\w–-])', True, '#1508p', 'conf_1e_1a_1es', 5, [['_c_conf_1e_1a_1es_1', '-', 'le', 0, 'Erreur de frappe\xa0?', ''], ['_c_conf_1e_1a_1es_2', '-', 'la', 0, 'Erreur de frappe\xa0?', ''], ['_c_conf_1e_1a_1es_3', '-', 'les', 0, 'Erreur de frappe\xa0?', '']]],
  ]],
  ["ocr", [
    ['(?i)(?<![\\w.,–-])[1[\\]][ea]s(?![\\w–-])', True, '#1520p', 'ocr_les', 3, [[None, '-', 'les', 0, 'Erreur de numérisation\xa0?', '']]],
    ["(?i)(?<![\\w.,–-])([1[][’'`‘])[aâeéèêëiîïoôuyh][\\w-]*(?![\\w–-])", True, '#1528p', 'ocr_l_regex', 3, [[None, '-', 'l’|L’|j’|J’', 1, 'Erreur de numérisation\xa0?', '']]],
    ['(?i)(?<![\\w.,–-])1fs?(?![\\w–-])', True, '#1536p', 'ocr_il_regex', 3, [['_c_ocr_il_regex_1', '-', 'Ils|ils|ifs', 0, 'Erreur de numérisation\xa0?', ''], ['_c_ocr_il_regex_2', '-', 'Il|il|if', 0, 'Erreur de numérisation\xa0?', '']]],
  ]],
  ["", [
    ['(?i)(?<![\\w.,–-])\\w+‑\\w+‑\\w+(?![\\w–-])', True, '#1553p', 'p_trait_union_conditionnel1', 4, [[None, '~', '=_p_p_trait_union_conditionnel1_1', 0]]],
    ['(?i)(?<![\\w.,–-])\\w+‑\\w+(?![\\w–-])', True, '#1554p', 'p_trait_union_conditionnel2', 4, [[None, '~', '=_p_p_trait_union_conditionnel2_1', 0]]],
    ['([?!…][?!… \xa0]*)[ "\'”»\xa0]*,', False, '#1557p', 'p_fin_dialogue', 4, [[None, '~', '*', 1]]],
  ]],
]

lSentenceRules = [
  ["", [
    ['(?<![\\w.,–-])(\\w+) {1,3}\\1(?![\\w–-])', False, '#1621s', 'doublon', 4, [['_c_doublon_1', '-', '\\1', 0, 'Doublon.', '']]],
  ]],
  ["num", [
    ['(?<![\\w.,–-])[\\dO]+[O][\\dO]+(?![\\w–-])', False, '#1635s', 'num_lettre_O_zéro1', 4, [['_c_num_lettre_O_zéro1_1', '-', '=_s_num_lettre_O_zéro1_1', 0, 'S’il s’agit d’un nombre, utilisez le chiffre «\xa00\xa0» plutôt que la lettre «\xa0O\xa0».', '']]],
    ['(?<![\\w.,–-])[1-9]O(?![\\w–-])', False, '#1636s', 'num_lettre_O_zéro2', 4, [['_c_num_lettre_O_zéro2_1', '-', '=_s_num_lettre_O_zéro2_1', 0, 'S’il s’agit d’un nombre, utilisez le chiffre «\xa00\xa0» plutôt que la lettre «\xa0O\xa0».', '']]],
  ]],
  ["", [
    ['[¹²³⁴⁵⁶⁷⁸⁹⁰]+', False, '#1646s', 'p_exposants', 4, [[None, '~', '*', 0]]],
    ['(?i)(?<![\\w.,–-])(\\w[\\w-]+)(\\d+)(?![\\w–-])', True, '#1650s', 'p_références_aux_notes', 4, [['_c_p_références_aux_notes_1', '~', '*', 2]]],
  ]],
  ["tu", [
    ['(?i)(?<![\\w.,–-])(\\w+)(?:--|—|–|−|⁃)(\\w+)(?![\\w–-])', True, '#1659s', 'tu_trait_union_douteux', 7, [['_c_tu_trait_union_douteux_1', '-', '\\1-\\2', 0, 'Trait d’union\xa0: un tiret simple suffit.', '']]],
    ["(?i)([-–—− ]t(?:[’' ][-–—−]?|[-–—−][’' ]?))(ie?ls?|elles?|on|tu)(?![\\w–-])", True, '#1667s', 'tu_t_euphonique_incorrect', 7, [['_c_tu_t_euphonique_incorrect_1', '-', '-', 1, 'Le “t” euphonique n’est pas nécessaire avec “\\2”.', 'http://bdl.oqlf.gouv.qc.ca/bdl/gabarit_bdl.asp?T1=t+euphonique&id=2513'], ['_c_tu_t_euphonique_incorrect_2', '-', '-t-', 1, 'Pour le “t” euphonique, il faut deux traits d’union. Pas d’apostrophe. Pas d’espace.', ''], ['_c_tu_t_euphonique_incorrect_3', '~', '-t-', 1]]],
    ["(?i)[td]([- ]t[-’' ])(?:ie?l|elle|on)(?![\\w–-])", True, '#1672s', 'tu_t_euphonique_superflu', 7, [[None, '-', '-', 1, 'Le “t” euphonique est superflu quand le verbe se termine par “t” ou “d”.', 'http://bdl.oqlf.gouv.qc.ca/bdl/gabarit_bdl.asp?T1=t+euphonique&id=2513'], ['_c_tu_t_euphonique_superflu_2', '~', '-t-', 1]]],
  ]],
  ["eleu", [
    ['(?i)[aec](-(ie?l|elle|on))(?![\\w–-])', True, '#1676s', 'eleu_t_euphonique_manquant', 4, [[None, '-', '-t-\\2', 1, 'Euphonie. Il faut un “t” euphonique.', 'http://bdl.oqlf.gouv.qc.ca/bdl/gabarit_bdl.asp?T1=t+euphonique&id=2513']]],
  ]],
  ["@@@@", [
    ['graphe0', '#1712s'],
    ['graphe1', '#2620s'],
    ['ocr', '#3914s'],
    ['graphe2', '#4873s'],
    ['intergraphe12a', '#19819s'],
    ['intergraphe12b', '#20214s'],
  ]],
  ["redon2", [
    ['(?i)(?<![\\w.,–-])(\\w[\\w-][\\w-]+\\w)[ ,].* (\\1)(?![\\w–-])', True, '#20823s', 'redondances_phrase', 4, [['_c_redondances_phrase_1', '-', '_', 2, 'Dans cette phrase, répétition de “\\1” (à gauche).', ''], ['_c_redondances_phrase_2', '-', '_', 1, 'Dans cette phrase, répétition de “\\1” (à droite).', '']]],
  ]],
  ["mc", [
    ['(?i)(?<![\\w.,–-])(\\w\\w+)-(\\w\\w+)(?![\\w–-])', True, '#20837s', 'mc_mot_composé', 4, [['_c_mc_mot_composé_1', '>', ''], ['_c_mc_mot_composé_2', '-', '\\1\\2', 0, 'Vous pouvez ôter le trait d’union.', ''], ['_c_mc_mot_composé_3', '-', '_', 0, 'Mot inconnu du dictionnaire.', '']]],
  ]],
  ["@@@@", [
    ['graphe3', '#20857s'],
    ['prégraphe_gv1', '#29340s'],
    ['verbes1', '#29570s'],
    ['verbes2', '#33331s'],
  ]],
]



"""
Grammar checker options manager
"""

# generated code, do not edit
# template: <gc_core/py/lang_core/gc_options.py>
# variables generated in <compile_rules.py>


import traceback


dOptions = {}

_sAppContext = "Python"


def load (sContext="Python"):
    "mandatory first function to call: variables initialization"
    global dOptions
    global _sAppContext
    _sAppContext = sContext
    dOptions = getDefaultOptions(sContext)


def setOption (sOpt, bVal):
    "set option <sOpt> with <bVal> if it exists"
    if sOpt in dOptions:
        dOptions[sOpt] = bVal


def setOptions (dOpt):
    "update the dictionary of options with <dOpt>, only known options are updated"
    for sKey, bVal in dOpt.items():
        if sKey in dOptions:
            dOptions[sKey] = bVal


def getOptions ():
    "return a copy of options as dictionary"
    return dOptions.copy()


def resetOptions ():
    "set options to default values"
    global dOptions
    dOptions = getDefaultOptions()


def displayOptions (sLang="fr"):
    "display the list of grammar checking options"
    print("Options:")
    print("\n".join( [ k+":\t"+str(v)+"\t"+getOptionLabels(sLang).get(k, ("?", ""))[0]  for k, v  in sorted(dOptions.items()) ] ))
    print("")


def getOptionLabels (sLang="${sLang}"):
    "returns dictionary of UI labels"
    if sLang in _dOptLabel:
        return _dOptLabel[sLang]
    return _dOptLabel["${sLang}"]


def getDefaultOptions (sContext=""):
    "returns dictionary of options"
    if not sContext:
        sContext = _sAppContext
    if sContext in _dDefaultOpt:
        return _dDefaultOpt[sContext].copy()    # duplication necessary, to be able to reset to default
    return _dDefaultOpt["Python"].copy()        # duplication necessary, to be able to reset to default


def getOptionsColors (sTheme="Default", sColorType="aRGB"):
    "returns dictionary of options colors"
    dOptColor = _dOptColor[sTheme]  if sTheme in _dOptColor  else  _dOptColor["Default"]
    dColorType = _dColorType[sColorType]  if sColorType in _dColorType  else _dColorType["aRGB"]
    try:
        return { sOpt: dColorType[sColor] for sOpt, sColor in dOptColor.items() }
    except KeyError:
        traceback.print_exc()
        return {}


lStructOpt = [['basic', [['typo', 'apos', 'eepi'], ['esp', 'tab'], ['nbsp', 'unit'], ['tu', 'maj', 'minis'], ['num', 'nf'], ['virg', 'poncfin'], ['ocr', 'chim'], ['liga', 'mapos']]], ['gramm', [['conf', 'loc', 'gn']]], ['verbs', [['infi', 'conj', 'ppas'], ['imp', 'inte', 'vmode']]], ['style', [['bs', 'pleo'], ['eleu', 'neg'], ['redon1', 'redon2']]], ['misc', [['date', 'mc']]], ['debug', [['idrule']]]]


_dDefaultOpt = {
    "Python": {'typo': True, 'apos': True, 'eepi': True, 'esp': True, 'tab': False, 'nbsp': True, 'tu': True, 'maj': True, 'minis': True, 'num': True, 'virg': True, 'poncfin': False, 'unit': True, 'nf': True, 'liga': False, 'mapos': False, 'chim': True, 'ocr': False, 'conf': True, 'loc': True, 'gn': True, 'infi': True, 'conj': True, 'ppas': True, 'imp': True, 'inte': True, 'vmode': True, 'bs': True, 'pleo': True, 'eleu': True, 'neg': False, 'redon1': False, 'redon2': False, 'date': True, 'mc': False, 'idrule': False, 'html': False, 'latex': True, 'md': True},
    "Server": {'typo': True, 'apos': True, 'eepi': True, 'esp': True, 'tab': True, 'nbsp': True, 'tu': True, 'maj': True, 'minis': True, 'num': True, 'virg': True, 'poncfin': False, 'unit': True, 'nf': True, 'liga': False, 'mapos': False, 'chim': True, 'ocr': False, 'conf': True, 'loc': True, 'gn': True, 'infi': True, 'conj': True, 'ppas': True, 'imp': True, 'inte': True, 'vmode': True, 'bs': True, 'pleo': True, 'eleu': True, 'neg': False, 'redon1': False, 'redon2': False, 'date': True, 'mc': False, 'idrule': False, 'html': False, 'latex': False, 'md': True},
    "Writer": {'typo': True, 'apos': True, 'eepi': True, 'esp': True, 'tab': False, 'nbsp': True, 'tu': True, 'maj': True, 'minis': True, 'num': True, 'virg': True, 'poncfin': False, 'unit': True, 'nf': True, 'liga': False, 'mapos': False, 'chim': False, 'ocr': False, 'conf': True, 'loc': True, 'gn': True, 'infi': True, 'conj': True, 'ppas': True, 'imp': True, 'inte': True, 'vmode': True, 'bs': True, 'pleo': True, 'eleu': True, 'neg': False, 'redon1': False, 'redon2': False, 'date': True, 'mc': False, 'idrule': False, 'html': False, 'latex': False, 'md': False}
}

_dColorType= {'aHSL': {'orange': [30, 70, 50], 'orange2': [40, 90, 50], 'violet': [270, 45, 35], 'bleu': [210, 50, 50], 'bleupr': [240, 40, 60], 'fuchsia': [300, 30, 40], 'cyan': [180, 50, 40], 'grey': [0, 50, 50], 'orange_vif': [30, 100, 60], 'jaune_vif': [55, 100, 40], 'violet_vif': [270, 100, 70], 'bleu_vif': [210, 100, 50], 'vert_vif': [120, 100, 40], 'cyan_vif': [180, 100, 40], 'bleupr_clair': [240, 70, 70]}, 'sCSS': {'orange': 'hsl(30, 70%, 50%)', 'orange2': 'hsl(40, 90%, 50%)', 'violet': 'hsl(270, 45%, 35%)', 'bleu': 'hsl(210, 50%, 50%)', 'bleupr': 'hsl(240, 40%, 60%)', 'fuchsia': 'hsl(300, 30%, 40%)', 'cyan': 'hsl(180, 50%, 40%)', 'grey': 'hsl(0, 50%, 50%)', 'orange_vif': 'hsl(30, 100%, 60%)', 'jaune_vif': 'hsl(55, 100%, 40%)', 'violet_vif': 'hsl(270, 100%, 70%)', 'bleu_vif': 'hsl(210, 100%, 50%)', 'vert_vif': 'hsl(120, 100%, 40%)', 'cyan_vif': 'hsl(180, 100%, 40%)', 'bleupr_clair': 'hsl(240, 70%, 70%)'}, 'aRGB': {'orange': [217, 128, 38], 'orange2': [242, 166, 13], 'violet': [89, 49, 129], 'bleu': [64, 127, 191], 'bleupr': [112, 112, 194], 'fuchsia': [133, 71, 133], 'cyan': [51, 153, 153], 'grey': [191, 64, 64], 'orange_vif': [255, 153, 51], 'jaune_vif': [204, 187, 0], 'violet_vif': [178, 102, 255], 'bleu_vif': [0, 127, 255], 'vert_vif': [0, 204, 0], 'cyan_vif': [0, 204, 204], 'bleupr_clair': [125, 125, 232]}, 'nInt': {'orange': 14254118, 'orange2': 15902221, 'violet': 5845377, 'bleu': 4227007, 'bleupr': 7368898, 'fuchsia': 8734597, 'cyan': 3381657, 'grey': 12533824, 'orange_vif': 16750899, 'jaune_vif': 13417216, 'violet_vif': 11691775, 'bleu_vif': 32767, 'vert_vif': 52224, 'cyan_vif': 52428, 'bleupr_clair': 8224232}}

_dOptColor = {'Default': {'typo': 'orange', 'apos': 'orange2', 'eepi': 'orange2', 'esp': 'orange', 'tab': 'orange', 'nbsp': 'orange', 'tu': 'orange', 'maj': 'orange', 'minis': 'orange', 'num': 'orange', 'virg': 'orange', 'poncfin': 'orange', 'unit': 'orange', 'nf': 'orange', 'liga': 'orange', 'mapos': 'orange', 'chim': 'orange', 'ocr': 'bleupr', 'conf': 'violet', 'loc': 'violet', 'gn': 'bleu', 'infi': 'fuchsia', 'conj': 'fuchsia', 'ppas': 'fuchsia', 'imp': 'fuchsia', 'inte': 'fuchsia', 'vmode': 'fuchsia', 'bs': 'cyan', 'pleo': 'cyan', 'eleu': 'cyan', 'neg': 'cyan', 'redon1': 'cyan', 'redon2': 'cyan', 'date': 'cyan', 'mc': 'cyan'}, 'Writer': {'typo': 'orange_vif', 'apos': 'jaune_vif', 'eepi': 'jaune_vif', 'esp': 'orange_vif', 'tab': 'orange_vif', 'nbsp': 'orange_vif', 'tu': 'orange_vif', 'maj': 'orange_vif', 'minis': 'orange_vif', 'num': 'orange_vif', 'virg': 'orange_vif', 'poncfin': 'orange_vif', 'unit': 'orange_vif', 'nf': 'orange_vif', 'liga': 'orange_vif', 'mapos': 'orange_vif', 'chim': 'orange_vif', 'ocr': 'bleupr_clair', 'conf': 'violet_vif', 'loc': 'violet_vif', 'gn': 'bleu_vif', 'infi': 'vert_vif', 'conj': 'vert_vif', 'ppas': 'vert_vif', 'imp': 'vert_vif', 'inte': 'vert_vif', 'vmode': 'vert_vif', 'bs': 'cyan_vif', 'pleo': 'cyan_vif', 'eleu': 'cyan_vif', 'neg': 'cyan_vif', 'redon1': 'cyan_vif', 'redon2': 'cyan_vif', 'date': 'cyan_vif', 'mc': 'cyan_vif'}}

_dOptLabel = {'fr': {'__optiontitle__': 'Grammalecte (Français)', 'basic': ['Typographie', ''], 'typo': ['Signes typographiques', ''], 'apos': ['Apostrophe typographique', 'Correction des apostrophes droites. Automatisme possible dans le menu Outils > Options d’autocorrection > Options linguistiques > Guillemets simples > Remplacer (à cocher)'], 'eepi': ['Écriture épicène', 'Normalisation de l’écriture épicène avec points médians.'], 'esp': ['Espaces surnuméraires', 'Signale les espaces inutiles entre les mots, en début et en fin de ligne.'], 'tab': ['Tabulations surnuméraires', 'Signale les tabulations inutiles en début et en fin de ligne.'], 'nbsp': ['Espaces insécables', 'Vérifie les espaces insécables avec les ponctuations «\xa0!\xa0?\xa0:\xa0;\xa0» (à désactiver si vous utilisez une police Graphite)'], 'maj': ['Majuscules', 'Vérifie l’utilisation des majuscules et des minuscules (par exemple, «\xa0la raison d’État\xa0», «\xa0les Européens\xa0»).'], 'minis': ['Majuscules pour ministères', 'Majuscules pour les intitulés des ministères.'], 'virg': ['Virgules', 'Virgules manquantes avant “mais”, “car” et “etc.”.'], 'poncfin': ['Ponctuation finale [!]', 'Vérifie s’il manque une ponctuation finale au paragraphe (seulement pour les paragraphes constitués de plusieurs phrases).'], 'tu': ['Traits d’union et soudures', 'Cherche les traits d’union manquants ou inutiles.'], 'num': ['Nombres', 'Espaces insécables sur les grands nombres (> 10 000). Vérifie la présence de «\xa0O\xa0» au lieu de «\xa00\xa0».'], 'unit': ['Espaces insécables avant unités de mesure', ''], 'nf': ['Normes françaises', ''], 'liga': ['Signaler ligatures typographiques', 'Ligatures de fi, fl, ff, ffi, ffl, ft, st.'], 'mapos': ['Apostrophe manquante après lettres isolées [!]', 'Apostrophe manquante après les lettres l d s n c j m t ç. Cette option sert surtout à repérer les défauts de numérisation des textes et est déconseillée pour les textes scientifiques.'], 'chim': ['Chimie', 'Typographie des composés chimiques (H₂O, CO₂, etc.).'], 'ocr': ['Erreurs de numérisation (OCR) [!]', 'Erreurs de reconnaissance optique des caractères. Beaucoup de faux positifs.'], 'gramm': ['Noms et adjectifs', ''], 'conf': ['Confusions et faux-amis', 'Cherche des erreurs souvent dues à l’homonymie (par exemple, les confusions entre «\xa0faîte\xa0» et «\xa0faite\xa0»).'], 'loc': ['Locutions', 'Écriture des locutions usuelles.'], 'gn': ['Accords (genre et nombre)', 'Accords des noms et des adjectifs.'], 'verbs': ['Verbes', ''], 'conj': ['Conjugaisons', 'Accord des verbes avec leur sujet.'], 'infi': ['Infinitif', 'Confusion entre l’infinitif et d’autres formes.'], 'imp': ['Impératif', 'Vérifie notamment la deuxième personne du singulier (par exemple, les erreurs\xa0: «\xa0vas…\xa0», «\xa0prend…\xa0», «\xa0manges…\xa0»).'], 'inte': ['Interrogatif', 'Vérifie les formes interrogatives et suggère de lier les pronoms personnels avec les verbes.'], 'ppas': ['Participes passés, adjectifs', ''], 'vmode': ['Modes verbaux', ''], 'style': ['Style', ''], 'bs': ['Populaire', 'Souligne un langage courant considéré comme erroné, comme «\xa0malgré que\xa0».'], 'pleo': ['Pléonasmes', 'Repère des redondances sémantiques, comme «\xa0au jour d’aujourd’hui\xa0», «\xa0monter en haut\xa0», etc.'], 'eleu': ['Élisions et euphonies', 'Signale les élisions incorrectes et les tournures dysphoniques.'], 'neg': ['Adverbe de négation [!]', 'Ne … pas, ne … jamais, etc.'], 'redon1': ['Répétitions dans le paragraphe [!]', 'Sont exclus les mots grammaticaux, ceux commençant par une majuscule, ainsi que “être” et “avoir”.'], 'redon2': ['Répétitions dans la phrase [!]', 'Sont exclus les mots grammaticaux, ainsi que “être” et “avoir”.'], 'misc': ['Divers', ''], 'mc': ['Mots composés [!]', 'Vérifie si les mots composés à trait d’union existent dans le dictionnaire (hormis ceux commençant par ex-, mi-, quasi-, semi-, non-, demi- et d’autres préfixes communs).'], 'date': ['Validité des dates', ''], 'debug': ['Débogage', ''], 'idrule': ['Identifiant des règles de contrôle [!]', 'Affiche l’identifiant de la règle de contrôle dans les messages d’erreur.']}, 'en': {'__optiontitle__': 'Grammar checking (French)', 'basic': ['Typography', ''], 'typo': ['Typographical glyphs', ''], 'apos': ['Typographical apostrophe', 'Detects typewriter apostrophes. You may get automatically typographical apostrophes in Tools > Autocorrect options > Localized options > Single quote > Replace (checkbox).'], 'eepi': ['Epicene writing', 'Normalization of epicene writing (usage of middle dots).'], 'esp': ['Useless spaces', 'Checks spaces within words and at the beginning and the end of lines.'], 'tab': ['Useless tabulations', 'Checks tabulations at the beginning and the end of lines.'], 'nbsp': ['Non-breakable spaces', 'Checks the use of non-breakable spaces with the following punctuation marks: «\xa0! ? : ;\xa0» (deactivate it if you use a Graphite font).'], 'maj': ['Capitals', 'Checks the use of uppercase and lowercase letters (i.e. «\xa0la raison d’État\xa0», «\xa0les Européens\xa0»).'], 'minis': ['Capitals for ministry', 'Capital letters for ministry.'], 'virg': ['Commas', 'Missing commas before “mais”, “car” and “etc.”.'], 'poncfin': ['Ending punctuation [!]', 'Checks if an ending punctuation is missing on paragraphs (only for multi-sentences paragraphs).'], 'tu': ['Hyphens and junctions', 'Checks missing or useless hyphens.'], 'num': ['Numerals', 'Large numerals and «\xa0O\xa0» instead of «\xa00\xa0».'], 'unit': ['Non-breaking spaces before units of measurement', ''], 'nf': ['French standards', ''], 'liga': ['Report typographical ligatures', 'Ligatures of fi, fl, ff, ffi, ffl, ft, st.'], 'mapos': ['Missing apostrophes after single letters [!]', 'Missing apostrophes after l d s n c j m t ç. This option is mostly useful to detect defects of digitized texts and is not recommended for scientific texts.'], 'chim': ['Chemistry', 'Typography for molecules (H₂O, CO₂, etc.)'], 'ocr': ['OCR errors [!]', 'Warning: many false positives.'], 'gramm': ['Nouns and Adjectives', ''], 'conf': ['Confusions and false friends', 'Seeks errors often due to homonymy (i.e. confusions between «\xa0faîte\xa0» et «\xa0faite\xa0»).'], 'loc': ['Locutions', 'Writing of common locutions.'], 'gn': ['Agreement (gender and number)', 'Agreement between nouns and adjectives.'], 'verbs': ['Verbs', ''], 'conj': ['Conjugation', 'Agreement between verbs and their subject.'], 'infi': ['Infinitive', 'Checks confusions between infinitive forms and other forms.'], 'imp': ['Imperative mood', 'Checks particularly verbs at second person singular (i.e. errors such as: «\xa0vas …\xa0», «\xa0prend …\xa0», «\xa0manges …\xa0»).'], 'inte': ['Interrogative mood', 'Checks interrogative forms and suggests linking the personal pronouns with verbs.'], 'ppas': ['Past participles, adjectives', 'Checks subject agreement with past participles and adjectives.'], 'vmode': ['Verbal modes', ''], 'style': ['Style', ''], 'bs': ['Popular style', 'Underlines misuse of language though informal and commonly used.'], 'pleo': ['Pleonasms', 'Semantic replications, like «\xa0au jour d’aujourd’hui\xa0», «\xa0monter en haut\xa0», etc.'], 'eleu': ['Élisions et euphonies', 'Signale les élisions incorrectes et les tournures dysphoniques.'], 'neg': ['Negation adverb [!]', 'Ne … pas, ne … jamais, etc.'], 'redon1': ['Duplicates in paragraph [!]', 'Are excluded grammatical words, words beginning by a capital letter, and also “être” and “avoir”.'], 'redon2': ['Duplicates in sentence [!]', 'Are excluded grammatical words, and also “être” and “avoir”.'], 'misc': ['Miscellaneous', ''], 'mc': ['Compound words [!]', 'Check if words with hyphen exist in the dictionary (except those beginning by ex-, mi-, quasi-, semi-, non-, demi- and other common prefixes).'], 'date': ['Date validity.', ''], 'debug': ['Debug', ''], 'idrule': ['Display control rule identifier [!]', 'Display control rule identifier in the context menu message.']}}
